"use strict";
this.name        = "exhibitions";
this.author      = "Norby";
this.copyright   = "2013 Norbert Nagy";
this.licence     = "CC BY-NC-SA 3.0";
this.description = "Ship exhibitions near witchpoints where player can earn Visitor Levels.";
this.version     = "1.20";

//customizable properties
this.$ExhibitionsAllShipsInPrivateExhibition = false; //show NPC ships in Private Exhibition (cheat)
this.$ExhibitionsLog = false; //verbose log
this.$ExhibitionsMaxShipsInLargeExhibitions = 144; //should be square number and min. 64
this.$ExhibitionsMinFPS = 20; //split Private Exhibition to get more than this frames per second
this.$ExhibitionsSpace = 150; //m space between ships in exhibition matrix

//internal properties, should not touch
this.$ExhibitionsEndDate = []; //store the last valid date of exhibitions
this.$ExhibitionsFCB = null; //FrameCallBack to fill up and rotate exhibition
this.$ExhibitionsRoles = //for large exhibitions, without escort, hunter, thargoid, trader, player and pirate
	["interceptor","miner","police","scavenger","shuttle","buoy","cargopod","missile"];
this.$ExhibitionsJ = 0; //index in player ships during exhibition creation, 0 mean no active exhibition
this.$ExhibitionsLastMenu = 0; //set marker to the lastly used exhibition menu line
this.$ExhibitionsLastStation = null; //Private Exhibition will before this station, update when dock into a new one
this.$ExhibitionsMenuFCB = null; //FrameCallBack for exhibitions menu
this.$ExhibitionsMO = null; //store initial orientation in menu
this.$ExhibitionsPerm = false; //permission to rotate ships in current exhibition
this.$ExhibitionsPermReqSent = false; //permission request sent
this.$ExhibitionsPosition = [null, null, null, null]; //pos, ori, vR, vU of exhibitions, will adjusted to player
this.$ExhibitionsSalon = 0; //Salon mean Private Exhibition in short, this variable avoid recreating ships after jump
this.$ExhibitionsSalonCost = 100; //must match with the cost of Private Exhibition Control in equipment.plist
this.$ExhibitionsSalonDelta = 0; //count delta to determine FPS in Private Exhibition
this.$ExhibitionsSalonDeltaC = 0; //counter to determine FPS in Private Exhibition
this.$ExhibitionsSalonJ = 0; //index in player ships during Private Exhibition creation
this.$ExhibitionsSalonO = null; //store Salon initial orientation
this.$ExhibitionsSalonP = null; //store Salon central position
this.$ExhibitionsSalonR = null; //store Salon initial vectorRight
this.$ExhibitionsSalonRot = true; //rotating Salon
this.$ExhibitionsSalonU = null; //store Salon initial vectorUp
this.$ExhibitionsSalonFCB = null; //FrameCallBack to fill up and rotate Salon
this.$ExhibitionsSalonShips = []; //store salon ships to fine if destroyed by player
this.$ExhibitionsSalonPart = 1; //actual part in splitted salon
this.$ExhibitionsSalonPartMax = 1; //salon splitted into how many parts
this.$ExhibitionsShips = []; //store ships to fine if destroyed by player
this.$ExhibitionsShipKeys = []; //store ship keys
this.$ExhibitionsSystems = []; //store system IDs of exhibitions
this.$ExhibitionsStartDate = []; //store start date of exhibitions
this.$ExhibitionsType = []; //store types of introduced exhibitions
this.$ExhibitionsTypes = [["All exhibitions","ship"], //0. line in this array mean all exhibitions
	["Large exhibitions","many"], //special type for $ExhibitionsRoles
	["Escort exhibitions","escort"], //type-role pairs, max. about 10 fit into the menu
	["Hunter exhibitions","hunter"],
	["Trader exhibitions","trader"],
	["Pirate meetings","pirate"]];
this.$ExhibitionsTimer = null; //wait for Gallery startUp finished
this.$ExhibitionsTimerPerm = null; //timer for permission
this.$ExhibitionsU = null; //store initial vectorUp
this.$ExhibitionsVisited = []; //store of the visited exhibitions
this.$ExhibitionsVisKey = []; //startdate+systemid of visited exhibitions (unique key)
this.$ExhibitionsVisitor = ["a Novice Visitor.", "an Unexpected Visitor.", "a Poor Visitor.",
	"a Below Average Visitor.", "an Average Visitor.", "an Above Average Visitor.",
	"a Competent Visitor.", "a Trustworthy Visitor!", "an Expected Visitor!",
	"the only Elite Visitor in this Ooniverse!"]; //string of visitor level


//equipment events
this.activated = function () { //attached to Private Exhibition Control equipment
	var e = worldScripts["exhibitions"];
	if( e.$ExhibitionsSalonRot ) {
		e.$ExhibitionsSalonRot = false;
		player.consoleMessage("Stop Private Exhibition rotation");
	} else {
		e.$ExhibitionsSalonRot = true;
		player.consoleMessage("Start Private Exhibition rotation");
	}
}

this.mode = function () { //attached to Private Exhibition Control equipment
	var e = worldScripts["exhibitions"];
	if( e.$ExhibitionsSalonPartMax > 1 ) { //step to the next part in splitted salon
		var p = e.$Exhibitions_SalonShips();
		if( e.$ExhibitionsSalonPart < e.$ExhibitionsSalonPartMax )
			e.$ExhibitionsSalonPart++; //next part
		else e.$ExhibitionsSalonPart = 1; //first part
		
		var s = e.$ExhibitionsSalonShips;
		for( var i = 0; i < s.length; i++ ) {
			if( s[i] && s[i].isValid ) s[i].remove(true); //remove other parts of salon
		}
		delete e.$ExhibitionsSalonShips;
		e.$ExhibitionsSalonShips = [];
		var sp = Math.floor( p.length / e.$ExhibitionsSalonPartMax ); ///ships in one part
		var imin = 0;
		if( e.$ExhibitionsSalonPart > 1 )
			imin = Math.floor( ( e.$ExhibitionsSalonPart - 1 ) * sp );
		var imax = p.length; //last part show last ship also
		if( e.$ExhibitionsSalonPart < e.$ExhibitionsSalonPartMax )
			imax = Math.floor( e.$ExhibitionsSalonPart * sp ) - 1;
		if( e.$ExhibitionsLog )
			log("Exhibitions", "Private Exhibition ships:"+p.length+" part:"+e.$ExhibitionsSalonPart
				+"/"+e.$ExhibitionsSalonPartMax+" sp:"+sp+" imin:"+imin+" imax:"+imax);
		for( var i = imin; i <= imax && i < p.length; i++ ) {
			e.$Exhibitions_Salon2( i );  //create ships in actual part
		}

		player.consoleMessage("Private Exhibit "+e.$ExhibitionsSalonPart //+"/"+e.$ExhibitionsSalonPartMax
			+": Showing "+(i-imin)+" of "+p.length+" ships");
	} else player.consoleMessage("Salon show "+e.$ExhibitionsSalonShips.length+" ships");
}

//world script events
this.startUp = function() {
	this.$ExhibitionsJ = 0; //index in player ships during exhibition creation
	this.$ExhibitionsLastMenu = 0; //set marker to the lastly used exhibition menu line
	if( player.ship && player.ship.docked )
		this.$ExhibitionsLastStation = player.ship.dockedStation;
	this.$ExhibitionsPerm = false; //permission to rotate ships in current exhibition
	this.$ExhibitionsPermReqSent = false; //permission request sent
	this.$ExhibitionsSalon = missionVariables.$ExhibitionsSalon; //recreate ships after load game if ordered before
	this.$ExhibitionsSalonDelta = 0; //count delta to determine FPS in Salon
	this.$ExhibitionsSalonDeltaC = 0; //counter to determine FPS in Salon
	this.$ExhibitionsSalonJ = 0; //index in player ships during Salon creation
	this.$ExhibitionsSalonRot = true; //rotating Salon
	this.$ExhibitionsSalonShips = []; //store ships to fine if destroyed by player
	this.$ExhibitionsSalonPartMax = 1;
	this.$ExhibitionsShips = []; //store ships to fine if destroyed by player
	this.$ExhibitionsSystems = []; //store system IDs of exhibitions

	var mv = missionVariables.$ExhibitionsSystems;//load advertised exhibitions from savegame
	if( !mv ) this.$ExhibitionsSystems = []; else this.$ExhibitionsSystems = mv.split(",");
	mv = missionVariables.$ExhibitionsStartDate;
	if( !mv ) this.$ExhibitionsStartDate = []; else this.$ExhibitionsStartDate = mv.split(",");
	mv = missionVariables.$ExhibitionsEndDate;
	if( !mv ) this.$ExhibitionsEndDate = []; else this.$ExhibitionsEndDate = mv.split(",");
	mv = missionVariables.$ExhibitionsType;
	if( !mv ) this.$ExhibitionsType = []; else this.$ExhibitionsType = mv.split(",");

	if( !(this.$ExhibitionsSystems.length > 0) ) //make new list if empty (first run)
		this.$Exhibitions_Create(clock.days);

	mv = missionVariables.$ExhibitionsVisited;
	if( !mv ) this.$ExhibitionsVisited = []; 
	else if( (mv+"").indexOf(",") ) this.$ExhibitionsVisited = (mv+"").split(","); //need +"" for bugfix
	else this.$ExhibitionsVisited = [mv];
	
	mv = missionVariables.$ExhibitionsVisKey;
	if( !mv ) this.$ExhibitionsVisKey = []; 
	else if( (mv+"").indexOf(",") ) this.$ExhibitionsVisKey = (mv+"").split(","); //need +"" for bugfix
	else this.$ExhibitionsVisKey = [mv];

	if( this.$ExhibitionsLog ) { //log first list also
		log("Exhibitions", "ExhibitionsSystems ("+this.$ExhibitionsSystems.length+"): "+this.$ExhibitionsSystems);
		log("Exhibitions", "ExhibitionsStartDate ("+this.$ExhibitionsStartDate.length+"): "+this.$ExhibitionsStartDate);
		log("Exhibitions", "ExhibitionsEndDate ("+this.$ExhibitionsEndDate.length+"): "+this.$ExhibitionsEndDate);
		log("Exhibitions", "ExhibitionsType ("+this.$ExhibitionsType.length+"): "+this.$ExhibitionsType);
		log("Exhibitions", "ExhibitionsVisited ("+this.$ExhibitionsVisited.length+"): "+this.$ExhibitionsVisited);
		log("Exhibitions", "ExhibitionsVisKey ("+this.$ExhibitionsVisKey.length+"): "+this.$ExhibitionsVisKey);
	}
	this.$ExhibitionsTimer = new Timer(this, this.$Exhibitions_Timed, 1);//need wait for gallery startUp finished

	if( player.ship && player.ship.docked ) this.shipDockedWithStation( player.ship.dockedStation ); //set interface
	else if( system ) this.shipDockedWithStation( system.mainStation );//fallback	
}

this.dayChanged = function(newday) {
	if(!system || system.isInterstellarSpace) return;
	this.$Exhibitions_Create(newday); //make new ones
	var openex = false;
	if( this.$ExhibitionsJ > 0 ) openex = true;
	if( this.$Exhibitions_CheckSystem() ) //create ships if there are an exhibition in this system today
		if( !openex ) player.consoleMessage("Exhibition open at witchpoint!",5);
	
	for( var x = 0; x < this.$ExhibitionsEndDate.length; x++) { //remove expired items from all array
		if( this.$ExhibitionsEndDate[ x ] < newday ) {
			this.$ExhibitionsSystems.splice(x,1);
			this.$ExhibitionsStartDate.splice(x,1);
			this.$ExhibitionsEndDate.splice(x,1);
			this.$ExhibitionsType.splice(x,1);
			x--;
		}
	}
}

this.playerBoughtEquipment = function(equipmentKey) {
	if(equipmentKey === ("EQ_SALON_CONTROL")) {
		this.$ExhibitionsSalon = 1; //recreate ships after load game
		clock.addSeconds(86400*(1+Math.random()));
		this.$Exhibitions_Salon();
	}
}

this.playerEnteredNewGalaxy = function(galaxyNumber) { //make new exhibitions
	this.$ExhibitionsEndDate = []; //store the last valid date of exhibitions
	this.$ExhibitionsSystems = []; //store system IDs of exhibitions
	this.$ExhibitionsStartDate = []; //store start date of exhibitions
	this.$ExhibitionsType = []; //store types of introduced exhibitions
	this.$Exhibitions_Create(clock.days); //recreate exhibitions
}

this.playerWillSaveGame = function(message) {
	missionVariables.$ExhibitionsVisited = this.$ExhibitionsVisited;
	missionVariables.$ExhibitionsVisKey = this.$ExhibitionsVisKey;
	missionVariables.$ExhibitionsSalon = this.$ExhibitionsSalon;
	missionVariables.$ExhibitionsSystems = this.$ExhibitionsSystems;
	missionVariables.$ExhibitionsStartDate = this.$ExhibitionsStartDate;
	missionVariables.$ExhibitionsEndDate = this.$ExhibitionsEndDate;
	missionVariables.$ExhibitionsType = this.$ExhibitionsType;
}

this.shipDockedWithStation = function(station)
{
	if( this.$ExhibitionsLastStation != station ) {
		this.$ExhibitionsLastStation = station; //save lastly docked station to check if changed
		this.$Exhibitions_SalonRemove();//to can buy new salon placed before this new station
	}
	if( isValidFrameCallback( this.$ExhibitionsFCB ) )
		removeFrameCallback( this.$ExhibitionsFCB );
		
	this.$ExhibitionsPerm = false; //permission to rotate ships in current exhibition
	this.$ExhibitionsPermReqSent = false; //permission request sent
	
	var len = 0;
	if( this.$ExhibitionsVisited ) len = this.$ExhibitionsVisited.length;
	var exs = len;
	var exv = "";
	if( len > 0 ) exv += " - you are "+(this.$ExhibitionsVisitor[ this.$Exhibitions_VisitorLevel() ]);
	
	station.setInterface("Exhibitions",{
		title: "Exhibitions ("+exs+" visited)"+exv,
		category: "Ship Systems",
		summary: "Check incoming ship meetings to extend your Gallery with new ships and increase your Visitor Level.",
		callback: this.$Exhibitions_Interface.bind(this)
		});
}

this.shipKilledOther = function(whom, damageType) {
	if( this.$ExhibitionsShips && this.$ExhibitionsShips.indexOf(whom) > -1 ) {
		player.consoleMessage("GalCop fined you "+formatCredits(10000, false, true)
			+" for destroyng an Exhibition ship", 10);
		player.credits -= 100000;
		player.score--;
	}
	if( this.$ExhibitionsSalonShips && this.$ExhibitionsSalonShips.indexOf(whom) > -1 ) {
		player.consoleMessage("GalCop fined you "+formatCredits(10000, false, true)
			+" for destroying a Private Exhibition ship", 10);
		player.credits -= 10000;
		player.score--;
	}
}

this.shipTargetAcquired = function(target) { //flag
	var e = worldScripts["exhibitions"];
	if( e.$ExhibitionsJ > 0 && e.$ExhibitionsShips && e.$ExhibitionsShips.indexOf(target) > -1 ) {
		var si = e.$ExhibitionsSystems.indexOf(system.ID+"");//need +"" for bugfix
		if( si > -1 ) {
		    var key = e.$ExhibitionsStartDate[si]+""+galaxyNumber+""+system.ID; //unique for each exhibitions
		    if( e.$ExhibitionsVisKey.indexOf(key) == -1 ) { //new visit
			var rep = e.$Exhibitions_VisitorLevel();
			if( !e.$ExhibitionsVisited ) this.$ExhibitionsVisited = []; //for sure
			if( !e.$ExhibitionsVisKey ) this.$ExhibitionsVisKey = []; //for sure
			var len = e.$ExhibitionsVisited.length;
			e.$ExhibitionsVisited.push(clock.days+": "+system.name+" in "+(galaxyNumber+1)+" galaxy");
			e.$ExhibitionsVisKey.push(key);
			var msg = "You visited your "+(len+1)+" exhibition.";
			player.consoleMessage(msg, 10);
			if( e.$ExhibitionsLog ) log("Exhibitions", msg+" "+key);
			var rep2 = e.$Exhibitions_VisitorLevel();
			if( rep2 > rep ) {
				msg = "Congratulations. You are "+e.$ExhibitionsVisitor[ rep2 ];
				player.commsMessage(msg, 10);
				if( e.$ExhibitionsLog ) log("Exhibitions", msg);
			}
		    }
		}
		//request permission for rotate ships
		if( !this.$ExhibitionsPerm ) {
			if( !this.$ExhibitionsTimerPerm && !this.$ExhibitionsPermReqSent ) {
				this.$ExhibitionsPermReqSent = true;
				player.consoleMessage("Permission requested to rotate ships in this exhibition", 10);
				var t = Math.ceil(Math.random() * 30); //max. 30 seconds
				this.$ExhibitionsTimerPerm = new Timer(this, this.$Exhibitions_TimedPerm, t);
			} else {
				player.consoleMessage("Please wait. You will get permission soon.", 5);
			}
		}
	}
}

this.shipWillEnterWitchspace = function() {
	delete this.$ExhibitionsShips;
	this.$ExhibitionsShips = [];
	if( isValidFrameCallback( this.$ExhibitionsFCB ) )
		removeFrameCallback( this.$ExhibitionsFCB );//stop rotate
	
	this.$ExhibitionsLastStation = null; //clear last station
	this.$ExhibitionsSalon = false; //do not recreate ships in salon after hyperjump
	this.$ExhibitionsSalonDeltaC = 0; //count delta to determine FPS in Salon
	delete this.$ExhibitionsSalonShips;
	this.$ExhibitionsSalonShips = [];
	this.$ExhibitionsSalonPartMax = 1;
	if( isValidFrameCallback( this.$ExhibitionsSalonFCB ) )
		removeFrameCallback( this.$ExhibitionsSalonFCB );//stop rotate
	player.ship.removeEquipment("EQ_SALON_CONTROL");
}

this.shipWillExitWitchspace = function() { //use this event due to shipExitedWitchspace is not working in v1.77
	this.$ExhibitionsPerm = false; //permission to rotate ships in current exhibition
	this.$ExhibitionsPermReqSent = false;
	if(system && !system.isInterstellarSpace) this.$Exhibitions_CheckSystem(); //create ships after hyperjump
}

this.shipWillLaunchFromStation = function() {
	player.ship.hudHidden = false;
	this.$ExhibitionsPerm = false; //permission to rotate ships in current exhibition
	this.$ExhibitionsPermReqSent = false; //permission request sent
	if( this.$ExhibitionsJ > 0 ) {
		if( !isValidFrameCallback( this.$ExhibitionsFCB ) )
			this.$ExhibitionsFCB = addFrameCallback( this.$Exhibitions_FCB ); //add ships and rotate
	} else this.$Exhibitions_CheckSystem(); //recreate ships after load game
}


//Exhibitions methods
this.$Exhibitions_CheckSystem = function() {
	var e = worldScripts["exhibitions"];
	var x = e.$ExhibitionsSystems.indexOf( system.ID+"" );//need +"" for bugfix
	if( e.$ExhibitionsLog ) log("Exhibitions", " CheckSystem:"+system.ID+" "+x+" "+e.$ExhibitionsJ+" "
		+clock.days+" "+e.$ExhibitionsStartDate[ x ]+"-"+e.$ExhibitionsEndDate[ x ]+" s"+e.$ExhibitionsSystems[x]);
	if( x != -1 && clock.days >= e.$ExhibitionsStartDate[ x ]
		&& clock.days <= e.$ExhibitionsEndDate[ x ] ) {
		if( e.$ExhibitionsJ == 0 ) e.$Exhibitions_Show( x ); //only if not shown before
		return( true );
	} else e.$Exhibitions_Remove(); //remove if last day elapsed
	return( false ); //no exhibition
}

this.$Exhibitions_Create = function( days ) {
	if(!system || system.isInterstellarSpace) return;
	
	var tlen = this.$ExhibitionsTypes.length; 
	var dlen = this.$ExhibitionsStartDate.length; 
	var maxstartdate = 0;
	if( dlen > 0 ) maxstartdate = this.$ExhibitionsStartDate.sort()[dlen-1];
	if( maxstartdate == 0 ) {
		maxstartdate = 2084002;
		if( galaxyNumber == 0 ) {
			this.$ExhibitionsSystems.push("55"); //long exhibition in Leesti for demonstration, need in "" for bugfix
			this.$ExhibitionsStartDate.push( 2084001 );
			this.$ExhibitionsEndDate.push( 2084060 );
			this.$ExhibitionsType.push( 1 ); //large exhibition
		}
	}
	var maxdate = days + 100; //new exhibitions introduced 100 days before
	if( !( dlen > 0 && maxstartdate > maxdate ) ) {
		for( var i = maxstartdate; i < maxdate; i++ ) {
			var r = system.scrambledPseudoRandomNumber(i);
			var t = Math.floor(r * (tlen)); //one open in every day, equal odds for each type
			var sy = Math.floor(r * 256) + ""; //need +"" for bugfix, 1/tlen odds for no opening
			if( t > 0 && this.$ExhibitionsSystems.indexOf(sy) == -1 ) {
				this.$ExhibitionsSystems.push( sy );
				this.$ExhibitionsStartDate.push( i );
				this.$ExhibitionsEndDate.push( i + 1 + Math.floor(r * 60) );//open max. 2 months
				this.$ExhibitionsType.push( t );
			}
		}
	}
}

this.$Exhibitions_FCB = function( delta ) {
	var e = worldScripts["exhibitions"];
	var p = e.$ExhibitionsShipKeys;
	if( e.$ExhibitionsJ < p.length ) {
		e.$Exhibitions_Show2( e.$ExhibitionsJ++ ); //spawn Salon ships in cycle
		if( e.$ExhibitionsJ >= p.length ) {
			var x = e.$ExhibitionsSystems.indexOf( system.ID+"" );//need +"" for bugfix
			var type = e.$ExhibitionsTypes[ e.$ExhibitionsType[ x ] ][1]; //escort, hunter, trader, pirate
			player.consoleMessage("Exhibition of "+type+" ships is open at witchpoint", 10);
		}
	} else {
		var t = worldScripts["towbar"];
		if( t && t.$TowbarShip && t.$TowbarShip.isValid && p.indexOf(t.$TowbarShip.dataKey) > -1 ) {
			player.consoleMessage("Towbar can not hold ships in Exhibition",10);
			t.$Towbar_Release();
		}
		
		if( e.$ExhibitionsPerm ) {
			var enc = worldScripts["gallery"].$GalleryEncounters;
			var s = e.$ExhibitionsShips;
			var m = null;
			for( var i = 0; i < s.length; i++ ) {
				m = s[i];
				if( enc.indexOf(m.dataKey) == -1 && m.orientation  //rotate new ship
					|| player.ship && player.ship.target == s[i] ) {//or current target
					m.orientation = m.orientation.rotateY( delta ).rotateZ( delta );
				}	
			}
		}
	}
}

this.$Exhibitions_Interface = function() {
	player.ship.hudHidden = true;//to shift down the menu
	this.$ExhibitionsLastMenu = 0;
	var g = worldScripts["gallery"];
	if( g && isValidFrameCallback( g.$GalleryFCB ) ) removeFrameCallback( g.$GalleryFCB ); //need for bugfix
	if( !isValidFrameCallback( this.$ExhibitionsMenuFCB ) )
		this.$ExhibitionsMenuFCB = addFrameCallback( this.$Exhibitions_MenuFCB ); //will call ex.menu
}

this.$Exhibitions_Menu = function() { //exhibitions menu
	var e = worldScripts["exhibitions"];
	var g = worldScripts["gallery"];
	var c = [];
	if(e.$ExhibitionsLog) log("Exhibitions", "Visitor:"+e.$Exhibitions_VisitorLevel()
		+" "+e.$ExhibitionsVisitor[e.$Exhibitions_VisitorLevel()]);

	var len = 0;
	if( e.$ExhibitionsVisited ) len = e.$ExhibitionsVisited.length;
	var exs = "You have not visited any exhibitions yet.";
	if( len > 0 ) {
		exs = "You visited "+len+" exhibition";
		if( len > 1 ) exs += "s";
		exs += ". You are "+(e.$ExhibitionsVisitor[ e.$Exhibitions_VisitorLevel() ]);
	}
	
	var msg = [];

	for( var i = 0; i < e.$ExhibitionsTypes.length; i++ ) { //make menu for each type, start from 1 (not from 0!)
		c.push(e.$ExhibitionsTypes[i][0]);
		var ex = "";
		var k = 0;
		for( var j = 0; j < e.$ExhibitionsType.length && k < 23 - e.$ExhibitionsTypes.length ; j++ ) {
			if( i == 0 || i == e.$ExhibitionsType[j] ) {//list one type into type specific menus
				k++;
				ex += e.$ExhibitionsStartDate[j] + " - " + e.$ExhibitionsEndDate[j] + "  "
					+ System.infoForSystem(galaxyNumber, e.$ExhibitionsSystems[j]).name;
				if( i == 0 ) ex += " - " + e.$ExhibitionsTypes[e.$ExhibitionsType[j]][1] + " ships\n";
				else ex += "\n";
					
			}
		}
		msg.push(ex);
	}
	var pur = "purchasable ships";
	if( e.$ExhibitionsAllShipsInPrivateExhibition ) pur = "all ships";
	c.push("Private Exhibition of "+pur);
	msg.push("You need "+formatCredits(e.$ExhibitionsSalonCost, false, true)+" to order your private exhibition.");
	c.push("Gallery of encounters"); 
	msg.push("Go to Gallery");
	c.push("Exit"); 
	msg.push("Bye");

	var startori = 0;
	var ti = "Exhibitions";
	if( e.$ExhibitionsLastMenu > 0 ) ti = c[e.$ExhibitionsLastMenu-10]+" "+clock.days+"-";
	else { e.$ExhibitionsMO = null; startori = 1; }
	msg[-10] = [ exs+"\n\nYou can choose exhibitions below. You will find them near witchpoints."
			+" Target at least one ship within to increase your Visitor Level."
			+" If you are lucky, you will find new ship types. You can add these to your Gallery by targeting them."
			+"\n\nIn the Private Exhibition menu, you can order an exhibition of "+pur+"."
			+"\nYou can go to the Gallery or exit when finished."];
	var ms = msg[e.$ExhibitionsLastMenu-10];
	if( e.$ExhibitionsLastMenu == 0 ) e.$ExhibitionsLastMenu = 10;
	var ch = {	"_10" : c[0],
		    	"_11" : c[1],
		    	"_12" : c[2],
		    	"_13" : c[3],
		    	"_14" : c[4],
		    	"_15" : c[5],
		    	"_16" : c[6],
		    	"_17" : c[7],
		    	"_18" : c[8],
		    	"_19" : c[9],
		    	"_20" : c[10],
		    	"_21" : c[21],
		    	"_22" : c[22],
		    	"_23" : c[23],
		    	"_24" : c[24],
		    	"_25" : c[25],
		    	"_26" : c[26],
		    	"_27" : c[27],
		    	"_28" : c[28],
		    	"_29" : c[29],
		    	"_30" : c[30],
		    	"_31" : c[31],
		    	"_32" : c[32],
		    	"_33" : c[33],
		    	"_34" : c[34],
		    	"_35" : c[35]
		};
	if( isValidFrameCallback( e.$ExhibitionsMenuFCB ) ) removeFrameCallback( e.$ExhibitionsMenuFCB );
	mission.runScreen({
		title: ti,
		message: ms,
		model: "["+player.ship.dataKey+"]",
		spinModel:false,
		background: "gallery_bg.png",
		initialChoicesKey: "_"+e.$ExhibitionsLastMenu,
		choices: ch
	},function(choice) {
		var ch = 10;
		if( choice ) ch = choice.substr(1); //cut starting "_"
		e.$ExhibitionsLastMenu = 1 * ch;
		if(e.$ExhibitionsLog) log("Exhibitions", "LastMenu: "+e.$ExhibitionsLastMenu);
		if( e.$ExhibitionsLastMenu - 10 < c.length - 3 ) {  //exhibition types, draw menu again
			if( !isValidFrameCallback( e.$ExhibitionsMenuFCB ) )
				e.$ExhibitionsMenuFCB = addFrameCallback( e.$Exhibitions_MenuFCB );
		} else switch( e.$ExhibitionsLastMenu - 10 ) {
			case (c.length - 3) : //Salon
				e.$Exhibitions_SalonMenu();
				break;
			case (c.length - 2) : //Go to Gallery
				g.$Gallery_Interface2( g.$GalleryShowAll ); //show from the lastly viewed ship
				break;
			case (c.length - 1) : //Exit
				break;
		}
	});
	var m = mission.displayModel;
	if( m ) {
		if( !e.$ExhibitionsMO ) e.$ExhibitionsMO = m.orientation; //save orientation in opening screen
		var w = oolite.gameSettings.gameWindow; //correction if not in widescreen
		var wide = Math.max( 0.01, ( w.width / w.height ) / g.$GalleryDefaultZoom );
		m.position = Vector3D(m.position.x, m.position.y, m.position.z * wide * 0.9); //zoom more closer
		m.orientation = e.$ExhibitionsMO.rotateZ( -Math.PI/2 ).rotateX( -Math.PI/2 )
			.rotateY( - Math.PI/4 * ( e.$ExhibitionsLastMenu + 5 - startori ) );
	}
}

this.$Exhibitions_MenuFCB = function( delta ) { //FrameCallBack for exhibitions menu
	var e = worldScripts["exhibitions"];
	if( isValidFrameCallback( e.$ExhibitionsMenuFCB ) ) removeFrameCallback( e.$ExhibitionsMenuFCB );
	e.$Exhibitions_Menu();
}


this.$Exhibitions_Remove = function() { //remove ships
	this.$ExhibitionsJ = 0; //index in ships during creation
	var s = this.$ExhibitionsShips;
	for( var i = 0; i < s.length; i++ ) {
		if( s[i] && s[i].isValid ) s[i].remove(true);
	}
	this.$ExhibitionsPosition = [null, null, null, null];
	if( isValidFrameCallback( this.$ExhibitionsFCB ) )
		removeFrameCallback( this.$ExhibitionsFCB );
}


this.$Exhibitions_Salon = function() {
	if(!system || system.isInterstellarSpace) return;

	this.$ExhibitionsSalonJ = 0; //index in player ships during Salon creation
	var st = player.ship; //at load game or in-fly debug
	if( player.ship.docked ) st = player.ship.dockedStation;
	this.$ExhibitionsSalonO = st.orientation;
	this.$ExhibitionsSalonP = st.position.add(st.vectorForward.multiply( 10300 )); //Salon central position
	this.$ExhibitionsSalonR = st.vectorRight; //store initial vectorRight
	this.$ExhibitionsSalonU = st.vectorUp; //store initial vectorUp
	player.ship.awardEquipment("EQ_SALON_CONTROL");//need after load game
	if( !isValidFrameCallback( this.$ExhibitionsSalonFCB ) )
		this.$ExhibitionsSalonFCB = addFrameCallback( this.$Exhibitions_SalonFCB );
}

this.$Exhibitions_Salon2 = function( j ) {
	if(!system || system.isInterstellarSpace) return;

	var e = worldScripts["exhibitions"];
	var g = worldScripts["gallery"];
	var p = e.$Exhibitions_SalonShips();
	var space = e.$ExhibitionsSpace;//between ships
	var st = player.ship; //at load game or in-fly debug
	if( player.ship.docked ) st = player.ship.dockedStation;
	var sq = Math.ceil(Math.sqrt(p.length));
//	for( var j = 0; j < p.length; j++ ) { //SalonFCB make this cycle to allow fast forwarding game clock parallelly
		var key = p[j];
		var row = Math.floor(j/sq);
		var pos = e.$ExhibitionsSalonP.add( e.$ExhibitionsSalonR.multiply( space * ( j - sq * row - sq / 2 + 0.5 ) ) )
				.add( e.$ExhibitionsSalonU.multiply( space * ( row - sq / 2 + 0.5 ) ) );//square align
		var ships = system.addShips("["+key+"]", 1, pos, 5);
		if( ( !ships || !ships[0] ) && key.indexOf("-player") > -1 ) { //handle unsuccesful ship creation
			key = key.slice(0, key.indexOf("-player")); //remove -player
			ships = system.addShips("["+key+"]", 1, pos, 5);
			if( ships ) e.$Exhibitions_SalonAdd( ships[0], key, st, e, space, j );
		} else if( ships && ships[0] ) e.$Exhibitions_SalonAdd( ships[0], key, st, e, space, j );
//	}
}

this.$Exhibitions_SalonAdd = function( ship, key, st, e, space, j ) {
	if( !ship ) { //handle unsuccesful ship creation
		if(e.$ExhibitionsLog) log("Exhibitions", "Failed addShips ["+key+"] to Private Exhibition.");
	} else {
		if(e.$ExhibitionsLog) log("Exhibitions", "["+key+"] added to Private Exhibition.");
		if( ship.escorts && ship.escorts.length > 0 ) 
			for( var i = 0; i < ship.escorts.length; i++ ) ship.escorts[i].remove(true);
		ship.bounty = 0;
		ship.displayName = "[Private Exhibition] " + ship.displayName;
		ship.setAI("nullAI.plist");
		ship.setScript("exhibitions-ship-script.js"); //must deactivate the original script
		ship.target = null;
		ship.clearDefenseTargets();
//		ship.scanClass = "CLASS_CARGO";
		if( st ) ship.orientation = e.$ExhibitionsSalonO.rotateX(3*Math.PI/2);
		var cr = ship.collisionRadius; //shift out large ships from the square plane
		if( cr > space ) {
			var c2 = 1; //prevent collision of two neighbour baseship
			if( Math.floor( j / 2 ) == j / 2 ) c2 = -1;
			var mul =  c2 * (cr + 4 * space);
			if(e.$ExhibitionsLog) log("Exhibitions", ship.displayName+" too large, shifted out "+j+" "+c2+" "+mul);
			ship.position = ship.position.add( st.vectorForward.multiply( mul ) );
		}
		e.$ExhibitionsSalonShips.push(ship);
	}
}

this.$Exhibitions_SalonFCB = function( delta ) {
	var e = worldScripts["exhibitions"];
	var p = e.$Exhibitions_SalonShips();
	if( e.$ExhibitionsSalonJ < p.length ) {
		if( e.$ExhibitionsSalonJ == 0 ) { //add active police as defenders of salon
			var role = "police";
			if( 6 <=  system.techLevel ) role = "interceptor";
			var eg = system.addShips(role, 1+Math.floor(system.government/2), e.$ExhibitionsSalonP, 2000);
			if( eg ) for( var i = 0; i < eg.length; i++ ) {
				if( eg[i] && eg[i].isValid ) eg[i].displayName = "[Private Exhibition Guard] "+eg[i].displayName;
			}
		}
		e.$Exhibitions_Salon2( e.$ExhibitionsSalonJ++ ); //spawn Salon ships in cycle
	} else {
		if( e.$ExhibitionsSalonDeltaC > 0 ) {
			if(e.$ExhibitionsLog) log("Exhibitions", "C:"+e.$ExhibitionsSalonDeltaC+" delta:"+delta
				+" deltasum:"+e.$ExhibitionsSalonDelta);
			e.$ExhibitionsSalonDeltaC++;
			e.$ExhibitionsSalonDelta += delta;
		}
		if( e.$ExhibitionsSalonJ == p.length ) {
			e.$ExhibitionsSalonJ++; //send message once only
			player.consoleMessage(p.length+" ships arrived. Take off to view Private Exhibition.",10);
			e.$ExhibitionsSalonRot = true; //rotating Salon
		}
		if( e.$ExhibitionsSalonJ > p.length && e.$ExhibitionsSalonDeltaC == 0
			&& player.ship && !player.ship.docked ) { //wait for undock
			e.$ExhibitionsSalonDelta = delta; //start count delta to determine FPS for Salon split
			e.$ExhibitionsSalonDeltaC = 1;
		}
		if( e.$ExhibitionsSalonPartMax == 1 && e.$ExhibitionsSalonDelta > 2 ) { //after 2 seconds of measurement
			var fps = e.$ExhibitionsSalonDeltaC / e.$ExhibitionsSalonDelta; ///frames per second
			e.$ExhibitionsSalonDelta = delta;
//			e.$ExhibitionsSalonDeltaC = 1;//restart measurement
			e.$ExhibitionsSalonDeltaC = 0;//stop measurement
			var split = "";
			if( fps < e.$ExhibitionsMinFPS //too few FPS
				&& player.ship && player.ship.position && e.$ExhibitionsSalonP
				&& player.ship.position.distanceTo(e.$ExhibitionsSalonP) < 25000 ) { //not so far
				e.$ExhibitionsSalonPartMax = Math.ceil( e.$ExhibitionsMinFPS / fps ); ///round up
				var sp = Math.floor( p.length / e.$ExhibitionsSalonPartMax ); ///ships in one part
				var sq = Math.ceil(Math.sqrt(p.length));
				if( sp < sq ) //at least one row of ships in one part
					e.$ExhibitionsSalonPartMax = Math.ceil( p.length / sq ); ///
				e.$ExhibitionsSalonPart = 0; //will show the first part
				split = ", split into "+e.$ExhibitionsSalonPartMax+" parts";
				var button = "b";
				if( oolite.gameSettings.keyConfig ) //from Oolite v.1.77.1
					button = String.fromCharCode( oolite.gameSettings.keyConfig.key_mode_equipment );
				player.consoleMessage("Private Exhibition split into "+e.$ExhibitionsSalonPartMax
					+" parts. Prime Private Exhibition Control and press '"+button+"' to show next part", 10);
				e.mode(); //hide parts of salon
			}
			if(e.$ExhibitionsLog)
				log("Exhibitions", "Private Exhibition "+p.length+" ships, "+fps+" FPS"+split);
		}

		var t = worldScripts["towbar"];
		if( t && t.$TowbarShip && t.$TowbarShip.isValid && p.indexOf(t.$TowbarShip.dataKey) > -1 ) {
			player.consoleMessage("Towbar can not hold ships in Private Exhibition",10);
			t.$Towbar_Release();
		}
		
		var s = e.$ExhibitionsSalonShips;
		var m = null;
		for( var i = 0; i < s.length; i++ ) {
			m = s[i];
			if( m && m.isValid && m.orientation && player.ship && //rotate all ships in salon
				( e.$ExhibitionsSalonRot && player.ship.target != s[i] //or current target only
				|| !e.$ExhibitionsSalonRot && player.ship.target == s[i] ) )
				m.orientation = m.orientation.rotateZ( delta/1.5 );
		}
	}
}

this.$Exhibitions_SalonMenu = function() {
	var e = worldScripts["exhibitions"];
	var g = worldScripts["gallery"];
	var pur = "purchasable ships";
	if( e.$ExhibitionsAllShipsInPrivateExhibition ) pur = "all ships";

	var msg = "You can order a Private Exhibition of "+pur
		+" before this station for "+formatCredits(e.$ExhibitionsSalonCost, false, true)
		+", but you need to allow 1-2 days for this to be set up. If you have an existing Private Exhibition,"
		+" it will be renewed with this purchase."
		+"\n\nYou will get a Private Exhibition Control primable equipment to start and stop rotation of ships"
		+" in your Private Exhibition, or you can start and stop rotation of a single ship by targeting it."
		+" Do not destroy ships in your Private Exhibition. If you do, you will be fined."
		+" Private Exhibition will be closed if you dock with another station or jump into another system.";
	mission.runScreen({
		title: "Order a Private Exhibition",
		message: msg,
		model: "["+player.ship.dataKey+"]",
		spinModel:false,
		background: "gallery_bg.png",
		choices: {"_0":"No thanks","_1":"OK. I will pay and wait for ships!"}
	},function(choice) {
		e.$ExhibitionsLastMenu = 0; //will display initial message
		switch(choice) {
			case "_0":
				if( !isValidFrameCallback( e.$ExhibitionsMenuFCB ) )
					e.$ExhibitionsMenuFCB = addFrameCallback( e.$Exhibitions_MenuFCB );
				break;
			case "_1":
				var saloncost = e.$ExhibitionsSalonCost; //must match with the cost in equipment.plist
				if( player.ship.equipmentStatus("EQ_SALON_CONTROL") == "EQUIPMENT_OK" ) {
					player.consoleMessage("Old Private Exhibition closed.",3);
					e.$Exhibitions_SalonRemove();
				}
				if( player.credits >= saloncost ) {
					player.credits -= saloncost;
					player.ship.awardEquipment("EQ_SALON_CONTROL");
					e.playerBoughtEquipment("EQ_SALON_CONTROL"); //do as if just bought one
					//no FCBE setup so will exit from ex.menu
				} else {
					//display not enough money message
					mission.runScreen({
						title: "Order a Private Exhibition",
						message: "You need "+formatCredits(e.$ExhibitionsSalonCost, false, true)
							+" to order your Private Exhibition.",
						model: "["+player.ship.dataKey+"]",
						spinModel:false,
						background: "gallery_bg.png",
						choices: {"_0":"Ok"}
						},function(choice) {
							//back to the exhibitions menu
							if( !isValidFrameCallback( e.$ExhibitionsMenuFCB ) )
							    e.$ExhibitionsMenuFCB = addFrameCallback(e.$Exhibitions_MenuFCB);
						});
				}
				break;
		}
	});
	var w = oolite.gameSettings.gameWindow;
	var wide = Math.max( 0.01, ( w.width / w.height ) / g.$GalleryDefaultZoom ); //correction if not in widescreen
	var m = mission.displayModel;
	if( m ) {
		m.position = Vector3D(m.position.x, m.position.y, m.position.z * wide * 0.9); //zoom more closer
		m.orientation = m.orientation.rotateZ( Math.PI/2 );
	}
}

this.$Exhibitions_SalonShips = function() {
	var g = worldScripts["gallery"];
	if( worldScripts["exhibitions"].$ExhibitionsAllShipsInPrivateExhibition ) 
		return( g.$GalleryAllObjects ); //cheat: all existing ships
	else return( g.$GalleryKeysForRole[ 1 ] );  //player ships
}

this.$Exhibitions_SalonRemove = function() {
	this.$ExhibitionsSalon = 0; //do not recreate ships after docked into a new station
	this.$ExhibitionsSalonDeltaC = 0; //count delta to determine FPS in Salon
	this.$ExhibitionsSalonJ = 0; //index in player ships during Salon creation
	this.$ExhibitionsSalonO = null; //store Salon initial orientation
	this.$ExhibitionsSalonP = null; //store Salon central position
	this.$ExhibitionsSalonR = null; //store Salon initial vectorRight
	this.$ExhibitionsSalonRot = true; //rotating Salon
	this.$ExhibitionsSalonU = null; //store Salon initial vectorUp
	var s = this.$ExhibitionsSalonShips;
	for( var i = 0; i < s.length; i++ ) {
		if( s[i] && s[i].isValid ) s[i].remove(true); //remove previous salon
	}
	delete this.$ExhibitionsSalonShips;
	this.$ExhibitionsSalonShips = [];
	this.$ExhibitionsSalonPartMax = 1;
	if( isValidFrameCallback( this.$ExhibitionsSalonFCB ) )
		removeFrameCallback( this.$ExhibitionsSalonFCB );
	player.ship.removeEquipment("EQ_SALON_CONTROL"); //to can buy again
}

this.$Exhibitions_Show = function( x ) { //create ships
	if(!system || system.isInterstellarSpace) return;

	var e = worldScripts["exhibitions"];
	var type = e.$ExhibitionsTypes[ e.$ExhibitionsType[ x ] ][1]; //escort, hunter, trader, pirate
	var enc = worldScripts["gallery"].$GalleryEncounters;
	e.$ExhibitionsShips = [];
	e.$ExhibitionsShipKeys = [];

	var sdk = [];
	if (0 < oolite.compareVersion("1.79")) { //slow and limited method, use before Oolite v1.79
		if( type == "many" ) { //from $ExhibitionsRoles
			for( var i = 0; i < e.$ExhibitionsRoles.length; i++ ) {
				var ships = system.addShips( e.$ExhibitionsRoles[i], //sum of 64 ship created
					Math.ceil( 64 / e.$ExhibitionsRoles.length ), [0,0,0], 1000000 );
				if( ships ) for( var i = 0; i < ships.length; i++ )
					if( ships[i] && ships[i].isValid && sdk.indexOf(ships[i].dataKey) == -1 )
						sdk.push( ships[i].dataKey );
			}
			if(e.$ExhibitionsLog) log("Exhibitions", type+": "+sdk);
			sdk = sdk.concat(enc); //player and encountered ships at the end, show if anothers is not enough
		} else {
			var ships = system.addShips(type, 64, [0,0,0], 1000000);
			if( ships ) for( var i = 0; i < ships.length; i++ )
				if( ships[i] && ships[i].isValid && sdk.indexOf(ships[i].dataKey) == -1 )
					sdk.push( ships[i].dataKey );
			if(e.$ExhibitionsLog) log("Exhibitions", type+": "+sdk);
		}
		if( !sdk || sdk.length < 1 ) { //fallback to player ships
			var ships = system.addShips("player", 64, [0,0,0], 1000000);
			if( ships ) for( var i = 0; i < ships.length; i++ )
				if( ships[i] && ships[i].isValid && sdk.indexOf(ships[i].dataKey) == -1 )
					sdk.push( ships[i].dataKey );
			if(e.$ExhibitionsLog) log("Exhibitions", "player: "+sdk);
		}
	} else { // Oolite v1.79 or later
		if( type == "many" ) { //from $ExhibitionsRoles
			for( var i = 0; i < e.$ExhibitionsRoles.length; i++ ) {
				sdk = sdk.concat(Ship.keysForRole( e.$ExhibitionsRoles[i] ));
			}
			if(e.$ExhibitionsLog) log("Exhibitions", type+": "+sdk);
			sdk = sdk.concat(enc); //player and encountered ships at the end, show if anothers is not enough
		} else {
			sdk = Ship.keysForRole( type );
			if(e.$ExhibitionsLog) log("Exhibitions", type+": "+sdk);
		}
		if( !sdk || sdk.length < 1 ) {
			sdk = Ship.keysForRole( "player" );//fallback to player ships
			if(e.$ExhibitionsLog) log("Exhibitions", "player: "+sdk);
		}
	}
	if( (!sdk || sdk.length < 1 ) && player.ship ) sdk = player.ship.dataKey;//last resort
	if( !sdk || sdk.length < 1 ) { //should not happed
		log("Exhibitions", "$Exhibitions_Show() failed. Can not find any dataKey to show");
		return; //give up and exit to prevent errors
	}

	var maxnewship = 2; //a few new ship added into every exhibition only else reserves run out too early
	var shipno = 12 + 52 * system.scrambledPseudoRandomNumber(e.$ExhibitionsType[ x ]); //max. 64 = 8*8
	if( type == "many" ) shipno = Math.max(64, system.pseudoRandomNumber * e.$ExhibitionsMaxShipsInLargeExhibitions);
	for( var i = 0; e.$ExhibitionsShipKeys.length < shipno && i < sdk.length; i++ ) {
		var key = sdk[i];
		if( ( key && enc.indexOf(key) != -1 || maxnewship-- > 0 ) 
			&& key.indexOf("station") == -1 && key.indexOf("rock-hermit") == -1 ) {
			//still not sure if this is a ship, further checking
			if( Ship.shipDataForKey ) { //new fast method in 1.79 since 2014.05.19
				var s = Ship.shipDataForKey(key);
				if(this.$ExhibitionsLog) log("Exhibitions", j+". "+key+" shipDataForKey "+s);
				if( s && s.max_flight_speed && s.max_flight_speed > 0 )
					e.$ExhibitionsShipKeys.push(key); //ok, add it
			} else { //old way for Oolite 1.77
				if( key.indexOf("spacebar") == -1 ) //must check every problematic key
					e.$ExhibitionsShipKeys.push(key); //maybe ok, add it
			}
		}
	}

	if( sdk && sdk.length < shipno ) //duplicate until contain enough ships
		while( e.$ExhibitionsShipKeys.length < shipno && e.$ExhibitionsShipKeys.length < 1000 )
			e.$ExhibitionsShipKeys = e.$ExhibitionsShipKeys.concat(e.$ExhibitionsShipKeys);
	e.$ExhibitionsShipKeys.splice( shipno ); //shrink before short, can appear ships from the end of the alphabet also
	e.$ExhibitionsShipKeys.sort();
	if(e.$ExhibitionsLog) log("Exhibitions", "ExhibitionsShipKeys: "+e.$ExhibitionsShipKeys);


	this.$ExhibitionsJ = 0; //index in player ships during exhibition creation
	if( !isValidFrameCallback( this.$ExhibitionsFCB ) )
		this.$ExhibitionsFCB = addFrameCallback( this.$Exhibitions_FCB ); //add ships and rotate

	//add active police/pirate as defenders of exhibition
	var role = "police";
	if( 6 <=  system.techLevel ) role = "interceptor";
	if(type == "pirate") role = "pirate";
	var base = 1;
	if(type == "many") base += 1;

	var pos = Vector3D(0,0,10000);
	var ori = Quaternion(0, 0, 0, 0);
	var vr = Vector3D(1, 0, 0);
	var vu = Vector3D(0, 1, 0);
	if( player.ship && player.ship.position ) {
		if( player.ship.position.distanceTo([0,0,0]) < 25600 ) 
			pos = player.ship.position.add(player.ship.vectorForward.multiply( 10000 ));
		ori = player.ship.orientation;
		vr = player.ship.vectorRight; //store initial vectorRight
		vu = player.ship.vectorUp; //store initial vectorUp
	}
	this.$ExhibitionsPosition = [pos, ori, vr, vu]; 
	
	var eg = system.addShips(role, base+Math.floor(system.government/4), pos, 2000);
	if( eg ) for( var i = 0; i < eg.length; i++ ) {
		if( eg[i] && eg[i].isValid ) eg[i].displayName = "[Exhibition Guard] "+eg[i].displayName;
	}
}

this.$Exhibitions_Show2 = function( j ) {
	if(!system || system.isInterstellarSpace) return;

	var e = worldScripts["exhibitions"];
	var p = e.$ExhibitionsShipKeys;
	var space = e.$ExhibitionsSpace;//between ships
	var sq = Math.ceil(Math.sqrt(p.length));
//	for( var j = 0; j < p.length; j++ ) { //FCB make this cycle
		var key = p[j];
		var row = Math.floor(j/sq);
		var pos = null;
		
		if( e.$ExhibitionsPosition && e.$ExhibitionsPosition[0] ) pos = e.$ExhibitionsPosition[0]
			.add( e.$ExhibitionsPosition[2].multiply( space * ( j - sq * row - sq / 2 + 0.5 ) ) )
			.add( e.$ExhibitionsPosition[3].multiply( space * ( row - sq / 2 + 0.5 ) ) );//square align
		else pos = [ space * ( j - sq * row - sq / 2 + 0.5 ), space * ( row - sq / 2 + 0.5 ), 10000 ];//fallback
		
		var ships = system.addShips("["+key+"]", 1, pos, 5);
		if( ( !ships || !ships[0] ) && key.indexOf("-player") > -1 ) { //handle unsuccesful ship creation
			key = key.slice(0, key.indexOf("-player")); //remove -player
			ships = system.addShips("["+key+"]", 1, pos, 5);
			if( ships ) e.$Exhibitions_ShowAdd( ships[0], key, e, space, j );
		} else if( ships && ships[0] ) e.$Exhibitions_ShowAdd( ships[0], key, e, space, j );
//	}
}

this.$Exhibitions_ShowAdd = function( ship, key, e, space, j ) {
	if( !ship ) { //handle unsuccesful ship creation
		if(e.$ExhibitionsLog) log("Exhibitions", "Failed addShips ["+key+"] to the Exhibition.");
	} else {
		if(e.$ExhibitionsLog) log("Exhibitions", "["+key+"] added to the Exhibition.");
		if( ship.escorts && ship.escorts.length > 0 ) 
			for( var i = 0; i < ship.escorts.length; i++ ) ship.escorts[i].remove(true);
		ship.bounty = 0;
		ship.displayName = "[Exhibition] " + ship.displayName;
		ship.setAI("nullAI.plist");
		ship.setScript("exhibitions-ship-script.js"); //must deactivate the original script
		ship.target = null;
		ship.clearDefenseTargets();
//		ship.scanClass = "CLASS_CARGO";
		var cr = ship.collisionRadius; //shift out large ships from the square plane
		if( cr > space ) {
			var c2 = 1; //prevent collision of two neighbour baseship
			if( Math.floor( j / 2 ) == j / 2 ) c2 = -1;
			var mul =  c2 * (cr + 4 * space);
			if(e.$ExhibitionsLog) log("Exhibitions", ship.displayName+" too large, shifted out "+j+" "+c2+" "+mul);
			ship.position = ship.position.add( [0, 0, mul ] );
		}
//		if(e.$ExhibitionsLog) log("Exhibitions", "playerori:"+player.ship.orientation);
		if( e.$ExhibitionsPosition && e.$ExhibitionsPosition[1] ) 
			ship.orientation = e.$ExhibitionsPosition[1].rotateX( Math.PI/2 );
		else if( player.ship && player.ship.orientation ) //fallback
			ship.orientation = player.ship.orientation.rotate( Vector3D(1,0,0), Math.PI/2 );

		if( !e.$ExhibitionsShips ) e.$ExhibitionsShips = [ship];
		else e.$ExhibitionsShips.push(ship);
	}
}

this.$Exhibitions_Timed = function() { //need wait for Gallery startUp finished to get filled ship dataKey arrays
	var e = worldScripts["exhibitions"];
	if( e.$ExhibitionsTimer ) {
		e.$ExhibitionsTimer.stop();
		delete e.$ExhibitionsTimer;
	}
	if(e.$ExhibitionsSalon) e.$Exhibitions_Salon(); //recreate after load game if ordered before
	e.$Exhibitions_CheckSystem(); //recreate exhibition after load game if any
}

this.$Exhibitions_TimedPerm = function() { //
	var e = worldScripts["exhibitions"];
	if( e.$ExhibitionsTimerPerm ) {
		e.$ExhibitionsTimerPerm.stop();
		delete e.$ExhibitionsTimerPerm;
	}
	if( e.$ExhibitionsPerm ) { //end of permitted time for rotation
		e.$ExhibitionsPerm = false;
		e.$ExhibitionsPermReqSent = false;
		player.consoleMessage("Permission revoked. You can request again by targeting another ship in the exhibition", 10);
	} else if( e.$ExhibitionsPermReqSent ) { //end of wait for permission
		e.$ExhibitionsPerm = true;
		e.$ExhibitionsPermReqSent = false;
		player.consoleMessage("Permission granted. You can rotate ships in the exhibition for 5 minutes", 10);
		e.$ExhibitionsTimerPerm = new Timer(e, e.$Exhibitions_TimedPerm, 300);
	}
}

this.$Exhibitions_VisitorLevel = function() {
	var e = worldScripts["exhibitions"];
	var len = 0;
	if( e.$ExhibitionsVisited ) len = e.$ExhibitionsVisited.length;
	var rep = 0;
	if( len < 2 ) rep = 0;
	else if( len < 4 ) rep = 1;
	else if( len < 8 ) rep = 2;
	else if( len < 16 ) rep = 3;
	else if( len < 32 ) rep = 4;
	else if( len < 64 ) rep = 5;
	else if( len < 100 ) rep = 6;
	else if( len < 300 ) rep = 7;
	else if( len < 1000 ) rep = 8;
	else rep = 9; //Elite Visitor
	return( rep );
}