"use strict";
this.name        = "gallery";
this.author      = "Norby";
this.copyright   = "2013 Norbert Nagy";
this.licence     = "CC BY-NC-SA 3.0";
this.description = "Show installed ships and objects in an interface screen.";
this.version     = "1.20";

//customizable properties
this.$GalleryAll = false; //Gallery of all objects (cheat)
this.$GalleryDefaultZoom = 1.5; //default size of ships
this.$GalleryMaxIt = 5; //max. iteration in Oolite v1.77, reduce if cause "Universe is full" or crash to desktop
this.$GalleryLog = false; //verbose log
this.$GalleryRoles = ["all","player", //main roles, you can add more but must leave these first two untouched
	"escort","hunter","interceptor","miner","pirate","police","trader","scavenger","shuttle","thargoid",
	"asteroid","boulder","buoy","cargopod","missile","station"];
this.$GallerySpeed = 2; //move and zoom speed

//internal properties, should not touch
this.$GalleryAllObjects = []; //store all ship and object keys
this.$GalleryAllNames = []; //all ship names for search
this.$GalleryEncounters = []; //store encountered ship keys for savegame also
this.$GalleryEncNames = []; //encountered ship names for search
this.$GalleryFCB = null; //FrameCallBack for rotating ship model
this.$GalleryFCB2 = null; //FrameCallBack for menu
this.$GalleryFCB3 = null; //FrameCallBack to fill up other role
this.$GalleryKey = null; //store last dataKey
this.$GalleryKeyi = []; //selected key indexes in the GalleryKeysForRole array
this.$GalleryKeysi = 0; //actual index in all Keys
this.$GalleryKeysForRole = []; //dataKeys for the selected role
this.$GalleryLastMenu = 0; //set marker to the lastly used menu line
this.$GalleryLastEnc = []; //lastly encountered dataKeys
this.$GalleryMore = true; //show more menu
this.$GalleryMove = 0; //move showed ship
this.$GalleryOri = null; //store last orientation
this.$GalleryOthers = [];
this.$GalleryPrevZ = 0; //store previous z position
this.$GalleryRole = 0; //selected role
this.$GalleryRotate = 0; //rotating showed ship around Y axis
this.$GalleryShift = 0; //store move multiplier calculated from ship size
this.$GalleryShiftX = 0; //store move position
this.$GalleryShiftY = 0; //store move position
this.$GalleryShiftZ = 0; //store move position
this.$GalleryShowAll = false; //show all ships and objects selected in search
this.$GalleryTimer = null; //help avoid timelimit
this.$GalleryTRole = 0; //actual role in timer
this.$GalleryTRI = 0; //Timer Role Iteration
this.$GalleryTrash = []; //store created ships to remove later
this.$GalleryZoom = 0; //zoom showed ship

//following arrays comes from Technical Reference Library OXP by spara
this.$GalleryClassifiedScanClasses = ["CLASS_MILITARY", "CLASS_POLICE", "CLASS_THARGOID"];
this.$GalleryClassifiedRoles = ["constrictor", "kephalan", "odonatean", "scorpax", "bigTrader", "monkpatrol", "griff_blackmonk_defenceship", "monkhit1", "monkhit2", "monkhold1", "monkhold2", "ev_green_gecko", "sin-pirate"];//classify constrictor, aliens oxp, bigTrader ships, black monks, green gecko oxp and capisastra oxp.


//world script events
this.startUp = function() {
	this.$GalleryAllObjects = []; //store all ship and object keys
	this.$GalleryAllNames = []; //all ship names for search
	var ge = missionVariables.$GalleryEncounters; //load encountered ships from savegame
	if( !ge ) this.$GalleryEncounters = []; else this.$GalleryEncounters = ge.split(",");//need at least 2 item exists
	if(this.$GalleryLog)
		log("Gallery", "Loaded Encounters ("+this.$GalleryEncounters.length+"): "+this.$GalleryEncounters);
	this.$GalleryEncNames = []; //encountered ship names for search
	this.$GalleryFCB = null; //FrameCallBack for rotating ship model
	this.$GalleryLastMenu = 0; //set marker to the lastly used menu line
	
	var ge = missionVariables.$GalleryLastEnc; //load order of encounters from savegame
	if( !ge ) this.$GalleryLastEnc = []; 
	else if( (ge+"").indexOf(",") ) this.$GalleryLastEnc = (ge+"").split(","); //need +"" for bugfix of a single item
	else this.$GalleryLastEnc = [ge];
	if(this.$GalleryLog)
		log("Gallery", "Loaded LastEnc ("+this.$GalleryLastEnc.length+"): "+this.$GalleryLastEnc);
		
	this.$GalleryMore = true; //show more menu
	this.$GalleryMove = 0; //move showed ship
	this.$GalleryRole = 0; //starting role in Roles array
	this.$GalleryRotate = 0; //rotating showed ship around Y axis
	this.$GallerySet = false; //settings menu
	this.$GalleryShiftX = 0; //store move position
	this.$GalleryShiftY = 0; //store move position
	this.$GalleryShiftZ = 0; //store move position
	this.$GalleryShowAll = false; //show all ships and objects selected in search
	this.$GalleryTrash = []; //store created ships to remove later
	this.$GalleryZoom = 0; //zoom showed ship
	
	if (0 < oolite.compareVersion("1.79")) { //slow and limited method, use before Oolite v1.79
//can cause long wait in startUp, "Universe is full" or crash to desktop before the spinning cobra if not enough memory
		log("Gallery", "Fallback to a slow and inaccurate method without Oolite v1.79");
		this.$GalleryKeysForRole[ 0 ] = [];
		this.$GalleryKeysForRole[ 0 ] = this.$GalleryKeysForRole[ 0 ].concat(this.$GalleryEncounters);
		var all = false;
		if( this.$GalleryAll || worldScripts["exhibitions"] 
			&& worldScripts["exhibitions"].$ExhibitionsAllShipsInPrivateExhibition ) all = true;
		for( var i = 1; i < this.$GalleryRoles.length && all || i == 1; i++ ) { //main roles
			var maxk = this.$GalleryMaxIt; //how many iteration max., reduce if cause "Universe is full" or crash to desktop
			var prevlen = 0;
			var num = 16; //how many ships created at one time
			if( i == 1 ) num = 64; //more if player role
			for( var k = 1; k <= maxk; k++ ) {
				if( this.$GalleryKeysForRole[ i ] ) prevlen = this.$GalleryKeysForRole[ i ].length;
				var ships = system.addShips(this.$GalleryRoles[i], num, [0,0,0], 1000000);
//				this.$GalleryTrash = this.$GalleryTrash.concat(ships);
				if(ships) for( var j = 0; j < ships.length; j++ ) { //created ships
					if( ships[j] ) {
						var key = ships[j].dataKey;
						if( !this.$GalleryKeysForRole[ i ] ) {
							this.$GalleryKeysForRole[ i ] = [key];
						} else if( this.$GalleryKeysForRole[ i ].indexOf(key) == -1 )
							this.$GalleryKeysForRole[ i ].push(key);//add new key
						ships[j].remove(true);//can not remove everything and reach Universe limit
						//all ship remove must be called with true parameter to avoid call shipDied
						//event which will drop the following error in the case of constrictor!
						//TypeError: killer is null in oolite-constrictor.js, line 72
//					if(this.$GalleryLog) log("Gallery", i + ". main role: " + this.$GalleryRoles[i]
//							+" "+k+". try, "+(j+1)+". key: "+key
//							+" GalleryKeysForRole["+i+"]:"+this.$GalleryKeysForRole[ i ]);
					}
				}
				if(this.$GalleryLog) log("Gallery", i + ". main role: " + this.$GalleryRoles[i]
					+" "+k+". try: "+this.$GalleryKeysForRole[ i ] );
				if( //k == 1 && this.$GalleryKeysForRole[ i ].length < num / 3 ||//too few object in this role
					prevlen == this.$GalleryKeysForRole[ i ].length) maxk = 0;//exit from cycle
			}
			if(this.$GalleryKeysForRole[ i ]) {
				this.$GalleryKeysForRole[ i ].sort();
				if(this.$GalleryLog) log("Gallery", i + ". role after "+(k-1)+" try: " + this.$GalleryRoles[i]
					+ " ("+this.$GalleryKeysForRole[ i ].length+"): " + this.$GalleryKeysForRole[ i ] );
				this.$GalleryKeysForRole[ 0 ] = this.$GalleryKeysForRole[ 0 ].concat(this.$GalleryKeysForRole[ i ]);
			}
		}
		this.$GalleryKeysForRole[ 0 ].sort();
		for( var i = 1; i < this.$GalleryKeysForRole[ 0 ].length; i++ ) { //remove duplicated keys
			if(this.$GalleryKeysForRole[0][i] == this.$GalleryKeysForRole[0][i-1]) {
				this.$GalleryKeysForRole[0].splice(i,1); //remove this item
				i--;
			}
		}
		if(this.$GalleryLog) log("Gallery", "All dataKeys ("+this.$GalleryKeysForRole[ 0 ].length
			+"): " + this.$GalleryKeysForRole[ 0 ] );

//		this.$GalleryTRole = 1;
//		this.$GalleryTRI = 1;
//		this.$GalleryTimer = new Timer(this, this.$Gallery_Timed, 1); //cycle in timer
		if( !this.$GalleryEncounters || this.$GalleryEncounters.length == 0 )
			this.$GalleryEncounters = this.$GalleryKeysForRole[ 1 ]; //store playable ships
		else {
			for( var i = 1; i < this.$GalleryKeysForRole[ 1 ].length; i++ ) {
				var key = this.$GalleryKeysForRole[ 1 ][i];
				if( this.$GalleryEncounters.indexOf( key ) == -1 )
					this.$GalleryEncounters.push( key );
			}
		}
	
	} else { //Oolite v1.79 or later
		var sdk = Ship.keysForRole( "player" );  //playable and encountered ships only
		if(this.$GalleryEncounters && this.$GalleryEncounters.length > 0) {
			for( var i = 1; i < this.$GalleryEncounters.length; i++ ) {
				var key = this.$GalleryEncounters[i];
				if( sdk.indexOf( key ) == -1 ) sdk.push( key );
			}
		}
		this.$GalleryEncounters = sdk;
		if( this.$GalleryAll ) var sdk = Ship.keys(); //all dataKeys

	    this.$GalleryKeysForRole[ 0 ] = sdk.sort(); //all or playable+encountered dataKeys
	    this.$GalleryKeyi[ 0 ] = 0;
	    if(this.$GalleryLog) log("Gallery", "All dataKeys ("+this.$GalleryKeysForRole[ 0 ].length+"): "
		    + this.$GalleryKeysForRole[ 0 ] );
	
	    for( var i = 1; i < this.$GalleryRoles.length; i++ ) { //main roles
		var roleKeys = Ship.keysForRole( this.$GalleryRoles[i] ).sort();
		
		if( !roleKeys || !roleKeys.length || roleKeys.length < 1 )
			this.$GalleryRoles.splice(i,1); //remove empty role
		else {

/* - validate check removed (too slow), check when try to show the item only
		    for( var j = 0; j < roleKeys.length; j++ ) {
			var key = roleKeys[j];
			var ships = system.addShips("["+key+"]", 1, [0,0,0], 25000);
			if( ( !ships || !ships[0] ) && key.indexOf("-player") > -1 ) { //handle unsuccesful ship creation
				key = key.slice(0, key.indexOf("-player")); //remove -player
				ships = system.addShips("["+key+"]", 1, [0,0,0], 25000);
				if( !ships || !ships[0] ) { //handle unsuccesful ship creation
					roleKeys.splice(j,1); //remove item
					j--;
					if(this.$GalleryLog) log("Gallery", "Failed addShips ["+key+"], removed.");
				} else ships[0].remove(true); //ok without -player
			} else {
				if( ships && ships[0] ) {
					var ship = ships[0];
					if( ship.roles ) for( var k = 0; k < ship.roles.length; k++ ) {
						if( ship.roles[k] && ship.roles[k].indexOf("subent") > -1 ) {
							//skip subentities in Griff_Shipset_Replace_v1.34.oxp
							roleKeys.splice(j,1); //remove item
							j--;
							k = ship.roles.length;//exit from cycle
							var msg = "Subentity ["+key+"] removed.";
							if(this.$GalleryLog) log("Gallery", msg);
							if(ship != player.ship) ship.remove(true);
						}
					}
					ship.remove(true); //ok
				} else if(this.$GalleryLog) log("Gallery", "Failed addShips ["+key+"], removed.");
			}
		    }
*/		
		    if( this.$GalleryAll ) this.$GalleryKeysForRole[ i ] = roleKeys;
		    else {
			this.$GalleryKeysForRole[ i ] = [];
			for( var k = 1; k < roleKeys.length; k++ ) {
				var key = roleKeys[k];
				if( this.$GalleryKeysForRole[ 0 ].indexOf( key ) != -1 )
					this.$GalleryKeysForRole[ i ].push( key ); //get playable or encountered keys only
			}
		    }
		    this.$GalleryKeyi[ i ] = 0;
		    if(this.$GalleryLog) log("Gallery", i + ". role: " + this.$GalleryRoles[i] 
			    + " ("+this.$GalleryKeysForRole[ i ].length+"): " + this.$GalleryKeysForRole[ i ] );
		}
	    }
	
	    if(this.$GalleryLog) var roles = Ship.roles(); //all roles
	    if(this.$GalleryLog) for( var j = 0; j < roles.length; j++ ) {
//		if( roles[j] && roles[j][0] != "[" ) { //many roles start with "["
			var roleKeys = Ship.keysForRole( roles[j] );
			if(this.$GalleryLog) log("Gallery", (j+1)+". "+roles[j]+" ("+roleKeys.length+"): "+roleKeys);
//		}
	    }
//	    if( !isValidFrameCallback( this.$GalleryFCB3 ) )
//		this.$GalleryFCB3 = addFrameCallback( this.$Gallery_OtherRole );//bug: give all keys, not others only

	}
	
	this.$GalleryAllObjects = this.$GalleryKeysForRole[ 0 ].sort(); //store all ships
//	this.$GalleryEncounters.sort();//Gallery_EncSort() is better after GalleryEncNames is filled

		
	//get encountered ship names
	for( var j = 0; j < this.$GalleryEncounters.length; j++ ) {
		var key = this.$GalleryEncounters[j];
		if( Ship.shipDataForKey ) { //new fast method in 1.79 since 2014.05.19
			var s = Ship.shipDataForKey(key);
			if(this.$GalleryLog) log("Gallery", j+". "+key+" shipDataForKey "+s);//debug
			if( s && s.name && s.name.length > 0 ) this.$Gallery_AddEncName(j, s.name); //check and store name
		} else { //old slower way to get ship names, need validate check also
			var ships = system.addShips("["+key+"]", 1, [0,0,0], 1000000);
			if( ( !ships || !ships[0] ) && key.indexOf("-player") > -1 ) { //handle unsuccesful ship creation
				key = key.slice(0, key.indexOf("-player")); //remove -player
				ships = system.addShips("["+key+"]", 1, [0,0,0], 1000000);
				if( !ships || !ships[0] ) { //handle unsuccesful ship creation
					this.$GalleryEncounters.splice(j,1); //remove item
					j--;
//					if(this.$GalleryLog) 
						log("Gallery", "Failed addShips ["+key+"], removed.");
				} else {
					this.$Gallery_AddEncName(j, ships[0].name); //check and store name
					ships[0].remove(true); //ok without -player
				}
			} else {
				if( ships && ships[0] ) {
					var ship = ships[0];
					if( ship.roles ) for( var k = 0; k < ship.roles.length; k++ ) {
						if( ship.roles[k] && ship.roles[k].indexOf("subent") > -1 ) {
							//skip subentities in Griff_Shipset_Replace_v1.34.oxp
							this.$GalleryEncounters.splice(j,1); //remove item
							j--;
							k = ship.roles.length;//exit from cycle
							var msg = "Subentity ["+key+"] removed.";
//							if(this.$GalleryLog) 
								log("Gallery", msg);
							if(ship != player.ship) ship.remove(true);
						}
					}
					this.$Gallery_AddEncName(j, ship.name); //check and store name
					ship.remove(true); //ok
				} else //if(this.$GalleryLog)
					log("Gallery", "Failed addShips ["+key+"], removed.");
			}
		}
	}

	if(this.$GalleryLog) log("Gallery", " Gallery ("+this.$GalleryEncounters.length+"): "+this.$GalleryEncounters);//debug
	this.$Gallery_EncSort();
	this.$GalleryKeysForRole[ 0 ] = this.$GalleryEncounters;//always start in enc. array
	if(this.$GalleryLog) log("Gallery", " Gallery ("+this.$GalleryEncounters.length+"): "+this.$GalleryEncounters);//debug

	if( player.ship && player.ship.docked ) this.shipDockedWithStation( player.ship.dockedStation ); //set interface
	else if( system ) this.shipDockedWithStation( system.mainStation );//fallback
}

this.commsMessageReceived = function(message, sender) {
	this.$Gallery_Encounter( sender );
}

this.distressMessageReceived = function(aggressor, sender) {
	this.$Gallery_Encounter( aggressor );
	this.$Gallery_Encounter( sender );
}
	
this.playerTargetedMissile = function(missile) {
	this.$Gallery_Encounter( missile );
}

this.playerWillSaveGame = function(message) {
	missionVariables.$GalleryEncounters = this.$GalleryEncounters;  //store encountered ships into savegame
	missionVariables.$GalleryLastEnc = this.$GalleryLastEnc;  //store order of encounters into savegame
}

this.shipAttackedOther = function(other) {
	this.$Gallery_Encounter( other );
}

this.shipAttackedWithMissile = function(missile, whom) {
	this.$Gallery_Encounter( missile );
	this.$Gallery_Encounter( whom );
}

this.shipBeingAttacked = function(whom) {
	this.$Gallery_Encounter( whom );
}

this.shipBeingAttackedUnsuccessfully = function(whom) {
	this.$Gallery_Encounter( whom );
}

this.shipCloseContact = function(otherShip) {
	this.$Gallery_Encounter( otherShip );
}

this.shipCollided = function(otherShip)
{
	this.$Gallery_Encounter( otherShip );
}

this.shipDockedWithStation = function(station)
{
	if( isValidFrameCallback( this.$GalleryFCB ) )
		removeFrameCallback( this.$GalleryFCB );
	var len = this.$GalleryLastEnc.length;
	var s = "";
	if( len > 1 ) s = "s";
	station.setInterface("Gallery",{
		title: "Gallery of encounters ("+this.$GalleryLastEnc.length+" item"+s+" stored)",
		category: "Ship Systems",
		summary: "Shows a gallery of ships and other space objects that you have encountered, with public technical data where available.",
		callback: this.$Gallery_Interface.bind(this)
		});
	if( this.$GalleryAll ) {
	    station.setInterface("GalleryAll",{
		title: "Gallery of all objects ("+this.$GalleryAllObjects.length+" objects)",
		category: "Ship Systems",
		summary: "Shows a gallery of all ships and other space objects, with public technical data where available.",
		callback: this.$Gallery_InterfaceAll.bind(this)
	    });
	}
}

this.shipEnteredStationAegis = function(station) {
	this.$Gallery_Encounter( station );
}

this.shipFiredMissile = function(missile, target) {
	this.$Gallery_Encounter( missile );
	this.$Gallery_Encounter( target );
}

this.shipKilledOther = function(whom, damageType) {
	this.$Gallery_Encounter( whom );
}

this.shipScoopedOther = function(whom) {
	this.$Gallery_Encounter( whom );
}

this.shipTakingDamage = function(amount, whom, type) {
	this.$Gallery_Encounter( whom );
}

this.shipTargetAcquired = function(target) {
	this.$Gallery_Encounter( target );
}

this.shipWillLaunchFromStation = function() {
	player.ship.hudHidden = false;
	if( isValidFrameCallback( worldScripts["gallery"].$GalleryFCB ) )
		removeFrameCallback( worldScripts["gallery"].$GalleryFCB );
}


//Gallery methods
this.$Gallery_AddEncName = function( j, n ) { //save name for search, check for empty and duplicated name
	if( n && n.length > 0 ) {
		var x = this.$GalleryEncNames.indexOf(n);
		if( x > -1 ) { //duplicated name, add number in parenthesis to the name
			for( var i = 1; i < 10000; i++ ) {
				var n2 = n + " (" + i + ")";
				if( this.$GalleryEncNames.indexOf(n2) == -1 ) i = 10000;//exit
			}
			this.$GalleryEncNames[j] = n2;
		} else this.$GalleryEncNames[j] = n;//save name
	} else this.$GalleryEncNames[j] = this.$GalleryEncounters[j];//datakey if name is empty
	if(this.$GalleryLog) log("Gallery", j+". "+this.$GalleryEncounters[j]+": "+this.$GalleryEncNames[j]);
}

this.$Gallery_Encounter = function( ship ) { //save the keys of encountered ships
	if( !ship || !ship.isValid || !player.ship || !player.ship.isValid ) return;
	var key = ship.dataKey;
	if( key && key != "telescopemarker" && key.indexOf("customshields") == -1
		&& ( !this.$GalleryEncounters || this.$GalleryEncounters.indexOf( key ) == -1 ) ) {
		if(this.$GalleryLog) log("Gallery", key+" encountered.");
		if( !this.$GalleryEncounters ) this.$GalleryEncounters = [ key ];
		else this.$GalleryEncounters.push( key ); //do not sort: EncNames array will be in wrong order!
		if( !this.$GalleryLastEnc ) this.$GalleryLastEnc = [ key ];
		else this.$GalleryLastEnc.push( key ); //do not sort at all: this array hold the order of encounters
		var j = this.$GalleryEncounters.length - 1; //last item
		this.$Gallery_AddEncName( j, ship.name ); //do not use displayName to avoid randomshipnames OXP
//		this.$GalleryEncNames.push( ship.name );
		player.consoleMessage(this.$GalleryEncNames[j]+" is now added to your ship's Gallery.",10); //as the "+(j+1)+". space object
		this.$Gallery_EncSort();//sort after consoleMessage only!
//		if( !this.$GalleryShowAll ) this.$GalleryKeysForRole[0] = this.$GalleryEncounters;//needed but EncSort do it


		if( this.$GalleryAll ) for( var i = 2; i < this.$GalleryRoles.length; i++ ) { //main roles
			if( ship.roles && ship.roles.indexOf( this.$GalleryRoles[i] ) != -1 ) {
				if( !this.$GalleryKeysForRole[ i ] ) this.$GalleryKeysForRole[ i ] = [key];
				else this.$GalleryKeysForRole[ i ].push( key );
				this.$GalleryKeysForRole[ i ].sort();
			}
		}
	}
}

this.$Gallery_EncSort = function() {
	var g = worldScripts["gallery"];
	var e = g.$GalleryEncounters;
	var n = g.$GalleryEncNames;
	var k = []; //dataKeys for names
	for( var i = 0; i < n.length; i++ ) {
		if( n[i] && n[i].length > 0 ) k[ n[i] ] = e[i];
		else {
			e[i] = null;
			n[i] = null;
		}
	}
	g.$GalleryEncNames.sort();
	n = g.$GalleryEncNames;
	var e2 = [];
	for( var i = 0; i < n.length; i++ ) e2[i] = k[ n[i] ];
	g.$GalleryEncounters = e2;
	if( !g.$GalleryShowAll ) g.$GalleryKeysForRole[ 0 ] = g.$GalleryEncounters;
}

this.$Gallery_FCB = function( delta ) { //FrameCallBack updating ship in sell salvage screen
	var m = mission.displayModel;
	if( m && m.isValid ) {
		var g = worldScripts["gallery"];
		
		switch( g.$GalleryRotate ) {
			case 0:
				m.orientation = m.orientation.rotateY( delta/1.5 );
				break;
			case 2:
				m.orientation = m.orientation.rotateY( -delta/1.5 );
				break;
			case 4:
				m.orientation = m.orientation.rotateX( delta/1.5 );
				break;
			case 6:
				m.orientation = m.orientation.rotateX( -delta/1.5 );
				break;
		}
		g.$GalleryOri = mission.displayModel.orientation;
		
		//Move and zoom, camera is at [0,0,0] facing [1,0,0,0]
		switch( g.$GalleryMove ) {
			case 1:
				g.$GalleryShiftX += g.$GallerySpeed * g.$GalleryShift * delta;
				break;
			case 3:
				g.$GalleryShiftX -= g.$GallerySpeed * g.$GalleryShift * delta;
				break;
			case 5:
				g.$GalleryShiftY += g.$GallerySpeed * g.$GalleryShift * delta;
				break;
			case 7:
				g.$GalleryShiftY -= g.$GallerySpeed * g.$GalleryShift * delta;
				break;
		}
		if( g.$GalleryZoom == 1 ) g.$GalleryShiftZ -= 3 * g.$GallerySpeed * g.$GalleryShift * delta;
		else if( g.$GalleryZoom == 3 ) g.$GalleryShiftZ += 3 * g.$GallerySpeed * g.$GalleryShift * delta;
		
		var w = oolite.gameSettings.gameWindow;
		var wide = Math.max( 0.01, ( w.width / w.height ) / g.$GalleryDefaultZoom ); //correction if not in widescreen
		mission.displayModel.position = Vector3D(g.$GalleryShiftX, g.$GalleryShiftY, g.$GalleryShiftZ * wide);
	}
}

this.$Gallery_FCB2 = function( delta ) { //FrameCallBack for menu
	var g = worldScripts["gallery"];
	if( isValidFrameCallback( g.$GalleryFCB2 ) ) removeFrameCallback( g.$GalleryFCB2 );
	g.$Gallery_Interface2( g.$GalleryShowAll );
}

this.Gallery_GetShipData = function(ship, all) {
	var equip="";
	var s="";
	s=ship.forwardWeapon;
	if( s!=null ) equip+="Forward "+s.name;
	var shipno = 0;
	var r =  worldScripts["rocketmenu"];
	if( r ) {
		shipno = r.$RocketMenu_Name.indexOf(ship.dataKey); //RocketShip with extra weapons
	}
	if( shipno > 0 ) {
		var pul = r.$RocketMenu_Pulse[shipno];
		if( pul > 0 ) {
			if( equip.length > 0 ) equip+="\n";
			if( pul > 1) equip+=pul+" ";
			equip+="Auto Pulse Laser";
			if( pul > 1) equip+="s";
		}
		var ham = r.$RocketMenu_Hammer[shipno];
		if( ham > 0 ) {
			if( equip.length > 0 ) equip+="\n";
			if( ham > 1) equip+=ham+" ";
			equip+="Hammer Laser";
			if( ham > 1) equip+="s";
		}
		var Sniper = r.$RocketMenu_Sniper[shipno];
		if( Sniper > 0 ) {
			if( equip.length > 0 ) equip+="\n";
			if( Sniper > 1) equip+=Sniper+" ";
			equip+="Sniper Laser";
			if( Sniper > 1) equip+="s";
		}
		var pha = r.$RocketMenu_Sapper[shipno];
		if( pha > 0 ) {
			if( equip.length > 0 ) equip+="\n";
			if( pha > 1) equip+=pha+" ";
			equip+="Sapper";
			if( pha > 1) equip+="s";
		}
	}
	s=ship.aftWeapon;
	if( s!=null ) {
		if( equip.length > 0 ) equip+="\n"; //Anaconda has aft weapon only
		equip+="Aft "+s.name;
	}
	s=ship.portWeapon;
	if( s!=null ) equip+="\nPort "+s.name;
	s=ship.starboardWeapon;
	if( s!=null ) equip+="\nStarboard "+s.name;
	if( shipno > 0 ) {
		var bt = r.$RocketMenu_Turrets[shipno];
		if( bt > 0) {  //RocketShip with turrets
			if( equip.length > 0 && !ship.portWeapon && !ship.starboardWeapon) equip+="\n";
			else equip+=", ";
			if(bt < 1) equip += (bt*10)+" Mini";
			else if(bt > 99) equip += Math.floor(bt/100)+" Strong";
			else equip += bt+" Ball";
			equip += " Turret";
			if( bt > 1) equip+="s";
		}
		if(ship.dataKey.indexOf("rocketcruiser") > -1) equip+=", 2 Main Turrets"; //rocketcruiser-player
	}
	if(equip.length > 0) equip+="\n";
	else if( ship.isPiloted || all ) equip+="No Laser\n";
	var eqs = "";
/*	var n = ""; //equipment list removed to make space
	var eq = ship.equipment;
	var i=-1;
	while(eq[++i]) {
		n = eq[i].name;
		if( n.length > 0 ) eqs += n + ", ";
		else {	//skip noname eqs awarded by NumericHUD to avoid timeLimit error
			for( var j = i + 10; j < eq.length; j+=10 ) {
				if( eq[j].name.length != 0 ) {
					i = j - 10; //end of skip
					j = eq.length;
				}
			}
			while( i < eq.length && eq[i].name.length == 0 ) i++;
			i--;
		}
	}
	eqs = eqs.substr( 0, eqs.length - 2 ); //cut last comma
	if( eqs.length > 0 ) equip+=eqs+"\n\n";
	else equip+="\n";//empty line before desc if no eqs
*/	var desc = "";
	if( ship.scriptInfo && ship.scriptInfo.buydesc )//show ship description if any
		desc="\""+ship.scriptInfo.buydesc+"\"\n";
	var sh = 0;
	if( ship == player.ship ) sh += ship.maxForwardShield;
	else {
		var w = worldScripts["NPC-shields"];
		if( w && ship.script ) {
			if( !ship.script.maxShieldStrength ) w.shipSpawned(ship); //add npc shield
			sh += ship.script.maxShieldStrength;
		}
		var w = worldScripts["customshields"];
		if( w && ship.script ) {
			if( !ship.script.customshieldsmaxforwardshieldlevel ) w.shipSpawned(ship); //add customshield
			sh += ship.script.customshieldsmaxforwardshieldlevel;
		}
	}
	if( ship.dataKey.indexOf("escape-capsule") > -1 ) sh = 0; //no shield on escape pod
	else if( worldScripts["shieldequalizercapacitors"] ) {
		if( ship.equipmentStatus("EQ_BIGSHCAP") == "EQUIPMENT_OK"
			&& ship.equipmentStatus("EQ_FORWARD_SHIELD_CAPACITOR") == "EQUIPMENT_OK" ) sh += 256;
		else if( ship.equipmentStatus("EQ_FORWARD_SHIELD_CAPACITOR") == "EQUIPMENT_OK" ) sh += 64;
	}
	var psd = "";//ship.displayName+
	var hd = "";
	if(!ship.hasHyperspaceMotor ) hd = "  No Hyperdrive";
	var speed = ship.maxSpeed;
	if( ship.script && ship.script.name === "rocketships" )
		speed = Math.round(ship.maxSpeed/3*2)+"+"+Math.round(ship.maxSpeed/3);
	if( speed > 0 ) psd += "Speed: "+speed+" mLS  Thrust: "+ship.maxThrust+"  ";
	if( all || speed > 0 && ship.isPiloted ) {
		psd += "\nPitch: "+Math.round(ship.maxPitch*100)/100+
			"  Roll: "+Math.round(ship.maxRoll*100)/100+
			"  Yaw: "+Math.round(ship.maxYaw*100)/100+hd+
//			"  Version: "+this.$shipVersion(ship)+
//			"  Service "+ship.serviceLevel+
			"\nCargo: "+ship.cargoSpaceCapacity;
		if( ship.extraCargo > 0 ) psd += "+"+parseInt(ship.extraCargo); //need Oolite v1.79
		else if( ship.scriptInfo && ship.scriptInfo.cargoext && parseInt(ship.scriptInfo.cargoext) > 0 )
			psd += "+"+parseInt(ship.scriptInfo.cargoext);
		psd += "t  ";
	}
	var sp = "";
	var sph = Math.round(ship.collisionRadius*ship.collisionRadius*Math.PI);
	if( sph < 1000000 ) sp = sph;
	else sp = Math.round(sph/100000)/10+"k";//base sphere too large to fit in the line so show in km^2

	var en = "";
	var r = "";
	if(all) en = "Energy: "+ship.maxEnergy;
	else {
		if( ship.isPiloted ) {
			var eb = Math.max(1, Math.floor(ship.maxEnergy/64));
			en = "EBank";
			if( eb > 1 ) en += "s";
			en += ": "+eb;
		}
	}

	if( ship.isPiloted || all ) {
		var er = 0;
		if( ship.energyRechargeRate > 0 ) er = parseInt(ship.energyRechargeRate); //need Oolite v1.79
		else if( ship.scriptInfo && ship.scriptInfo.recharge && parseInt(ship.scriptInfo.recharge) > 0 )
			er = parseInt(ship.scriptInfo.recharge);
		if( er > 0 ) {
			var ers = "";
			if( er < 2.5 ) ers = "Poor";
			else if( er < 3.5 ) ers = "Medium";
			else if( er < 4.5 ) ers = "Good";
			else if( er < 10 ) ers = "Excellent";
			else ers = "Extreme";
			if( all ) ers += " ("+er+")";
			r = "  Recharge: "+ers;
		}
		if( sh > 0 ) r += "\nShields: "+Math.max(2, Math.floor(sh/64));
		if( ship.missileCapacity > 0 ) {
			if( sh > 0 ) r += "  "; else r += "\n";
			r += "Missile";
			if( ship.missileCapacity > 1 ) r += "s";
			r += ": "+ship.missileCapacity;
		}
	}

	psd += "Mass: "+Math.max(Math.round(ship.mass/1000), 1)+ //min.1t
		"t\nSize: "+Math.round(ship.boundingBox.x)+"*"+
		Math.round(ship.boundingBox.y)+"*"+Math.round(ship.boundingBox.z)+
		"m Radius: "+Math.round(ship.collisionRadius)+"m"+//" Sphere: "+sp+"m^2"+
		"\n"+en+r;
//	if( ship.scriptInfo && ship.scriptInfo.hardarmour && parseInt(ship.scriptInfo.hardarmour) > 0 )
//		psd += "\nHard Armour deflect "+parseInt(ship.scriptInfo.hardarmour)+" points from any damage";
		//removed, redundant with Hard Armour Equipment and give more free space
	psd += "\n"+equip;
	if( worldScripts["gallery"].$GalleryMore ) psd = desc + psd;
	if( ship.price > 0 ) psd += "\nPrice: "+formatCredits(ship.price, false, true);
	return(psd);
}

this.$Gallery_Interface = function() { //encounters only
	player.ship.hudHidden = true;//to shift down the menu
	var g = worldScripts["gallery"];
	g.$GalleryShowAll = false;
	g.$GalleryKeysForRole[ 0 ] = g.$GalleryEncounters; //fill up default key array
	
	var e = worldScripts["exhibitions"];
	if( e && isValidFrameCallback( e.$ExhibitionsMenuFCB ) )
		removeFrameCallback( e.$ExhibitionsMenuFCB );  //need for bugfix
	
	this.$Gallery_LastEncounters( false, true ); //not all, init
//	this.$Gallery_Search( false, true ); //not all, init
//	this.$Gallery_Interface2( false );
}

this.$Gallery_InterfaceAll = function() { //all ships
	player.ship.hudHidden = true;//to shift down the menu
	var g = worldScripts["gallery"];
	g.$GalleryShowAll = true;
	g.$GalleryKeysForRole[ 0 ] = g.$GalleryAllObjects; //fill up default key array
	this.$Gallery_Search( true, true ); //all, init
//	this.$Gallery_Interface2( true );
}

this.$Gallery_Interface2 = function( all ) {
	if( player.ship && player.ship.docked && system ) {
		player.ship.hudHidden = true;//to shift down the menu
		var g = worldScripts["gallery"];
		if( !g.$GalleryKeyi[g.$GalleryRole] ) g.$GalleryKeyi[g.$GalleryRole] = 0; //prevent a bug
		var key = g.$GalleryKeysForRole[g.$GalleryRole][g.$GalleryKeyi[g.$GalleryRole]];
		if(!key) {
			g.$GalleryKeyi[g.$GalleryRole] = 0;
			key = g.$GalleryKeysForRole[g.$GalleryRole][g.$GalleryKeyi[g.$GalleryRole]];
			if(!key) {
				g.$GalleryRole = 0;
				key = g.$GalleryKeysForRole[g.$GalleryRole][g.$GalleryKeyi[g.$GalleryRole]];
				if(!key) {
					g.$GalleryKeyi[g.$GalleryRole] = 0;
					key = g.$GalleryKeysForRole[g.$GalleryRole][g.$GalleryKeyi[g.$GalleryRole]];
					if(!key) return; //no key found
				}
			}
		}
		if( key == player.ship.dataKey ) var ship = player.ship;
		else {
			if( !all && 
				( g.$GalleryPlayerShips && g.$GalleryPlayerShips.indexOf(key) == -1 ) &&
				( g.$GalleryEncounters && g.$GalleryEncounters.indexOf(key) == -1 ) ) {
				if(this.$GalleryLog) log("Gallery", key+" not encountered so removed. ");
				this.$Gallery_RemoveCurrentShip(g, all); //will retry also
				return;
			}
			var ships = system.addShips("["+key+"]", 1, [0,0,0], 25000);
			if( !ships ) { //need to handle unsuccesful ship creation
				var p = key.indexOf("-player");
				if( p > -1 ) var key = key.slice(0, p); //remove -player part
				ships = system.addShips("["+key+"]", 1, [0,0,0], 25000);
				if( !ships ) { //need to handle unsuccesful ship creation
					var msg = "Failed addShips with role ["+key+"], removed.";
					if(this.$GalleryLog) log("Gallery", msg);
					this.$Gallery_RemoveCurrentShip(g, all); //will retry also
					return;
				}
			}
			var ship = ships[0];
			if( ship.roles ) for( var i = 0; i < ship.roles.length; i++ ) {
				if( ship.roles[i] && ship.roles[i].indexOf("subent") > -1 ) {
					i = ship.roles.length; //found, will exit from cycle
					//skip subentities in Griff_Shipset_Replace_v1.34.oxp
					var msg = "Subentity ["+key+"] removed.";
					if(this.$GalleryLog) log("Gallery", msg);
					if(ship != player.ship) ship.remove(true);
					this.$Gallery_RemoveCurrentShip(g, all); //will retry also
					return;
				}
			}
		}
		if(this.$GalleryLog) log("Gallery", "dataKey: "+key+ " Keyi:"+g.$GalleryKeyi[g.$GalleryRole]+
			" Gallery ("+g.$GalleryKeysForRole.length+"): "+g.$GalleryKeysForRole );

		var c = [key,
			"Previous",
			"Role: "+g.$GalleryRoles[g.$GalleryRole]+
				" ("+g.$GalleryKeysForRole[g.$GalleryRole].length+")",
			"Previous role",
			"Search",
			["Rotate Y+","Rotate stop","Rotate Y-","Rotate stop","Rotate X+","Rotate stop","Rotate X-","Rotate stop"],
			["Move stop","Move X+","Move stop","Move X-","Move stop","Move Y+","Move stop","Move Y-"],
			["Zoom stop","Zoom +","Zoom stop","Zoom -"],
			"Exit",
			"Hide Menu",
			"Menu"];
			
		if( !all && !g.$GalleryValidateTarget(ship) ) {
			var msg = "No public details.";
			if(this.$GalleryLog) log("Gallery", key+" showed without data.");
		} else {
			var msg = g.Gallery_GetShipData(ship, all);
			if(this.$GalleryLog) log("Gallery", msg);
		}
		if(ship != player.ship) ship.remove(true); //will really disappear after the end of the function
		
		var title = ship.displayName;
		var x = g.$GalleryEncounters.indexOf(key);
		if( x > -1 ) title = g.$GalleryEncNames[x]; //show (1), (2), etc. after duplicated names
		if( all ) name = key;
		else name = title;
		if( !name || name.length < 1) name = "..."; //for sure
		
		var ch = {	"_0" : name,
				"_10" : c[10]
		};
		if( g.$GalleryMore ) {
			if( all ) ch = {	"_0" : name,
		    			"_1" : c[1],
		    			"_2" : c[2],
		    			"_3" : c[3],
		    			"_4" : c[4],
		    			"_5" : c[5][g.$GalleryRotate],
		    			"_6" : c[6][g.$GalleryMove],
		    			"_7" : c[7][g.$GalleryZoom],
		    			"_8" : c[8],
		    			"_9" : c[9]
				};
			else ch = {	"_0" : name,
		    			"_1" : c[1],
//		    			"_2" : c[2], //without role menus
//		    			"_3" : c[3],
		    			"_4" : c[4],
		    			"_5" : c[5][g.$GalleryRotate],
		    			"_6" : c[6][g.$GalleryMove],
		    			"_7" : c[7][g.$GalleryZoom],
		    			"_8" : c[8],
		    			"_9" : c[9]
				};
		} 
		mission.runScreen({
			title: title,
			message: msg,
			model: "["+key+"]",
			spinModel:false,
			background: "gallery_bg.png",
			initialChoicesKey: "_"+g.$GalleryLastMenu,
			choices: ch
		},function(choice) {
			switch(choice) {
				case "_0":
					if( ++g.$GalleryKeyi[g.$GalleryRole] >= g.$GalleryKeysForRole[g.$GalleryRole].length
						|| !g.$GalleryKeysForRole[g.$GalleryRole][g.$GalleryKeyi[g.$GalleryRole]] ) 
						g.$GalleryKeyi[g.$GalleryRole] = 0;
					g.$GalleryLastMenu = 0;
					if( !isValidFrameCallback( g.$GalleryFCB2 ) )
						g.$GalleryFCB2 = addFrameCallback( g.$Gallery_FCB2 );
				break;
				case "_1":
					if( --g.$GalleryKeyi[g.$GalleryRole] < 0 )
						g.$GalleryKeyi[g.$GalleryRole] = 
							g.$GalleryKeysForRole[g.$GalleryRole].length - 1;
					g.$GalleryLastMenu = 1;
					if( !isValidFrameCallback( g.$GalleryFCB2 ) )
						g.$GalleryFCB2 = addFrameCallback( g.$Gallery_FCB2 );
				break;
				case "_2":
					if( ++g.$GalleryRole >= g.$GalleryRoles.length )
//						|| !g.$GalleryKeysForRole[g.$GalleryRole][g.$GalleryKeyi[g.$GalleryRole]] )
						g.$GalleryRole = 0;
					while( g.$GalleryRole > 0 &&
						!g.$GalleryKeysForRole[g.$GalleryRole]
						|| g.$GalleryKeysForRole[g.$GalleryRole].length < 1 ) { //skip empty roles
						if( ++g.$GalleryRole >= g.$GalleryRoles.length ) g.$GalleryRole = 0;
					}
					g.$GalleryLastMenu = 2;
					if( !isValidFrameCallback( g.$GalleryFCB2 ) )
						g.$GalleryFCB2 = addFrameCallback( g.$Gallery_FCB2 );
				break;
				case "_3":
					if( --g.$GalleryRole < 0 ) g.$GalleryRole =  g.$GalleryRoles.length - 1;
					while( g.$GalleryRole > 0 &&
						!g.$GalleryKeysForRole[g.$GalleryRole]
						|| g.$GalleryKeysForRole[g.$GalleryRole].length < 1 ) { //skip empty roles
						g.$GalleryRole--;
					}
					g.$GalleryLastMenu = 3;
					if( !isValidFrameCallback( g.$GalleryFCB2 ) )
						g.$GalleryFCB2 = addFrameCallback( g.$Gallery_FCB2 );
				break;
				case "_4":
					g.$GalleryLastMenu = 0;
					g.$Gallery_Search( all, false ); //use textEntry from Oolite v1.79
				break;
				case "_5":
					if( ++g.$GalleryRotate >= c[5].length ) g.$GalleryRotate = 0;
					g.$GalleryLastMenu = 5;
					mission.displayModel.orientation = g.$GalleryOri;
					if( !isValidFrameCallback( g.$GalleryFCB2 ) )
						g.$GalleryFCB2 = addFrameCallback( g.$Gallery_FCB2 );
				break;
				case "_6":
					if( ++g.$GalleryMove >= c[6].length ) g.$GalleryMove = 0;
					g.$GalleryLastMenu = 6;
					if( !isValidFrameCallback( g.$GalleryFCB2 ) )
						g.$GalleryFCB2 = addFrameCallback( g.$Gallery_FCB2 );
				break;
				case "_7":
					if( ++g.$GalleryZoom >= c[7].length ) g.$GalleryZoom = 0;
					g.$GalleryLastMenu = 7;
					if( !isValidFrameCallback( g.$GalleryFCB2 ) )
						g.$GalleryFCB2 = addFrameCallback( g.$Gallery_FCB2 );
				break;
				case "_8": //exit
				break;
				case "_9": //Hide Menu
					g.$GalleryMore = false;
					g.$GalleryLastMenu = 10;
					if( !isValidFrameCallback( g.$GalleryFCB2 ) )
						g.$GalleryFCB2 = addFrameCallback( g.$Gallery_FCB2 );
				break;
				case "_10": //Menu
					g.$GalleryMore = true;
					g.$GalleryLastMenu = 9;
					if( !isValidFrameCallback( g.$GalleryFCB2 ) )
						g.$GalleryFCB2 = addFrameCallback( g.$Gallery_FCB2 );
				break;
			}
		});
		if( g.$GalleryLog ) log("Gallery", key +" "+ g.$GalleryKey);
		if( key != g.$GalleryKey ) { //save position if new model
			g.$GalleryKey = key;
			g.$GalleryShiftX = mission.displayModel.position.x;
			g.$GalleryShiftY = mission.displayModel.position.y;
			if( g.$GalleryPrevZ < 1 || g.$GalleryShiftZ < 1 )
				g.$GalleryShiftZ = mission.displayModel.position.z;
			else {
				g.$GalleryShiftZ = mission.displayModel.position.z * g.$GalleryShiftZ / g.$GalleryPrevZ; ///
				g.$GalleryPrevZ = mission.displayModel.position.z;
			}
//			g.$GalleryOri = mission.displayModel.orientation;
		} else mission.displayModel.orientation = g.$GalleryOri;
		
		g.$GalleryShift = Math.max( ship.collisionRadius * 0.7 ,  //fix for adder and moray
					0.5 * mission.displayModel.position.z
					- Math.max( ship.boundingBox.x, ship.boundingBox.y ) ) / 5;
		if( g.$GalleryLog ) log("Gallery", mission.displayModel.position+ " shift:"+ g.$GalleryShift);
		// camera is at [0,0,0] facing [1,0,0,0]
		if( !isValidFrameCallback( g.$GalleryFCB ) )
			g.$GalleryFCB = addFrameCallback( g.$Gallery_FCB );
 		g.$Gallery_FCB(0);//needed to set position immediately
	}
}

this.$Gallery_LastEncounters = function( all, init ) {
	var g = worldScripts["gallery"];
//	g.$GalleryLastEnc = g.$GalleryEncounters;//debug

	var bgkey = player.ship.dataKey;
	var c = [];
	var title = g.$Gallery_SearchHeadTitle( all );
	if( !g.$GalleryLastEnc || g.$GalleryLastEnc.length == 0 ) { //skip last encounters menu if none
		var msg = "You have not found any object yet, but you can search in purchasable ships.";
//		g.$Gallery_Search( false, true ); //not all, init
	} else {
		var len = g.$GalleryLastEnc.length;
		var s = "";
		if( len > 1 ) s = "s";
		var msg = "You have so far encountered a total of "+len+" space object"+s
			+". Below are the latest items added to your gallery. Select the gallery item that you wish to view.";
//		var msg = g.$Gallery_SearchHeadMsg( all )+"   Last encounters:";
		var i = len - 1;
		bgkey = g.$GalleryLastEnc[ i ];
		for( var j = 0; i >= 0 && j < 33; j++ ) { //in reverse order
			var ei = g.$GalleryEncounters.indexOf( g.$GalleryLastEnc[ i ] );
			if( ei > -1 ) {
				c[ j ] = g.$GalleryEncNames[ ei ]; 
			} else j--;
			i--;
		}
	}
	
		var ch = {
			"_10" : c[0],
		    	"_11" : c[1],
		    	"_12" : c[2],
		    	"_13" : c[3],
		    	"_14" : c[4],
		    	"_15" : c[5],
		    	"_16" : c[6],
		    	"_17" : c[7],
		    	"_18" : c[8],
		    	"_19" : c[9],
		    	"_20" : c[10],
		    	"_21" : c[21],
		    	"_22" : c[22],
		    	"_23" : c[23],
		    	"_24" : c[24],
		    	"_25" : c[25],
		    	"_26" : c[26],
		    	"_27" : c[27],
		    	"_28" : c[28],
		    	"_29" : c[29],
		    	"_30" : c[30],
		    	"_31" : c[31],
		    	"_32" : c[32],
			"_S" : "Search" //max. 33 llines fit into the screen after two line msg and one empty line
		};
		
		if( isValidFrameCallback( worldScripts["gallery"].$GalleryFCB ) )
			removeFrameCallback( worldScripts["gallery"].$GalleryFCB );
		mission.runScreen({
			title: title,
			message: msg,
			background: "gallery_bg.png",
			model: "["+bgkey+"]",
			spinModel:false,
//			initialChoicesKey: "_"+g.$GalleryLastMenu,
			choices: ch
		},function( choice ) {
			g.$GalleryRole = 0;
			if( choice == "_S" ) {
				g.$Gallery_Search( all, false ); //use textEntry from Oolite v1.79
				return;
			}
			var text = "";
			if( choice ) text = c[ choice.substr(1) - 10 ]; //cut starting "_"
			g.$Gallery_SearchKey(text.toLowerCase(), all);
		});
		var m = mission.displayModel;
		if( m ) {
			var w = oolite.gameSettings.gameWindow; //correction if not in widescreen
			var wide = Math.max( 0.01, ( w.width / w.height ) / g.$GalleryDefaultZoom );
			m.position = Vector3D(m.position.x, m.position.y, m.position.z * wide );
		}
//	}
}

this.$Gallery_OtherRole = function() { //separated from startUp for faster boot
	var g = worldScripts["gallery"];
	var k = g.$GalleryKeysForRole[ 0 ];  //do one check/frame = 60 check/sec, need about 10 sec to finish
	var i = g.$GalleryKeysi++;
	var o = g.$GalleryOthers;
	
//	var k = Ship.keys(); //all dataKeys
//	var o = [];
//	for( var i = 0; i < k.length; i++ ) { //find keys not in main rules
		var ki = k[i];
		var ok = true;
		for( var j = 0; j < g.$GalleryKeysForRole.length; j++ ) {
			if( g.$GalleryKeysForRole.indexOf( ki ) > -1 ) { //bug: can not filter anything
				ok = false;
				j = g.$GalleryKeysForRole.length; //exit from cycle
			}
		}
		if( ok ) {//found new key
//			var key = ki;
//			var ships = system.addShips("["+key+"]", 1, [0,0,0], 25000);
//			if( ( !ships || !ships[0] ) && key.indexOf("-player") > -1 ) { //handle unsuccesful ship creation
//				key = key.slice(0, key.indexOf("-player")); //remove -player
//				ships = system.addShips("["+key+"]", 1, [0,0,0], 25000);
//				if( !ships || !ships[0] ) { //handle unsuccesful ship creation
//					if(g.$GalleryLog) log("Gallery", "Failed addShips ["+key+"], skipped.");
//				} else {  //ok without -player
//					ships[0].remove(true);
//					o.push( ki );//add to others
//				}
//			} else {
//				if( ships && ships[0] ) { //ok
//					ships[0].remove(true);
					o.push( ki );//add to others
//				} else if(g.$GalleryLog) log("Gallery", "Failed addShips ["+key+"], skipped.");
//			}
		}
//	}
	i++;
	if( i >=  k.length ) {
		o.sort();
		if( g.$GalleryLog) log("Gallery", "All other dataKeys: " + o );
		if( o.length > 0 ) {
			g.$GalleryKeysForRole.push( o );
			g.$GalleryRoles.push("all others");
		}
		if( isValidFrameCallback( g.$GalleryFCB3 ) )
			removeFrameCallback( g.$GalleryFCB3 );
	}
	g.$GalleryOthers = o;
}

this.$Gallery_RemoveCurrentShip = function(g, all) {
	g.$GalleryKeysForRole[g.$GalleryRole].splice(g.$GalleryKeyi[g.$GalleryRole],1); //remove from the current list
	if(!g.$GalleryKeysForRole[g.$GalleryRole]
		|| g.$GalleryKeysForRole[g.$GalleryRole].length < 1 ) //none left in this rule
		g.$GalleryRole = 0;
	else if(g.$GalleryKeyi[g.$GalleryRole] > g.$GalleryKeysForRole[g.$GalleryRole].length)
		g.$GalleryKeyi[g.$GalleryRole] ==
			g.$GalleryKeysForRole[g.$GalleryRole].length - 1; //previous one
	g.$Gallery_Interface2( all ); //try again with another ship
}
	
this.$Gallery_Search = function( all, init ) {
	var g = worldScripts["gallery"];
	var title = this.$Gallery_SearchHeadTitle( all );
	var msg = this.$Gallery_SearchHeadMsg( all );

	if (init || 0 < oolite.compareVersion("1.79")) { //no textEntry before Oolite v1.79
		msg += " Select the gallery item that you wish to view.";

		var k = [];
		if( g.$GalleryShowAll ) k = g.$GalleryKeysForRole[ 0 ];
		else k = g.$GalleryEncNames;
		var k1 = [];
		var l = ["a","b","c","d","e","f","g","h","i","j","k","l","m","n","o",
			"p","q","r","s","t","u","v","w","x","y","z"];
		var z = [];
		for( var j = 0; j < k.length; j++ ) 
			if(k[j]) k1[j] = k[j].charAt(0).toLowerCase(); 
			else if(g.$GalleryKeysForRole[ 0 ][j]) k1[j] = g.$GalleryKeysForRole[ 0 ][j].charAt(0).toLowerCase();
			else k1[j] = "?";
		for( var i = 0; i < l.length; i++ ) {
			var x = k1.indexOf( l[i] );
			if( x != -1 ) z[ l[i] ] = k[x];
		}
		var last = "";
		if( g.$GalleryLastEnc && g.$GalleryLastEnc.length > 0 ) last = "Last encounters";
		var ch = {	"_" : last,
				"A" : z.a,
				"B" : z.b,
				"C" : z.c,
				"D" : z.d,
				"E" : z.e,
				"F" : z.f,
				"G" : z.g,
				"H" : z.h,
				"I" : z.i,
				"J" : z.j,
				"K" : z.k,
				"L" : z.l,
				"M" : z.m,
				"N" : z.n,
				"O" : z.o,
				"P" : z.p,
//				"Q" : z.q,//removed to give a line to the exhibitions menu
				"R" : z.r,
				"S" : z.s,
				"T" : z.t,
				"U" : z.u,
				"V" : z.v,
				"W" : z.w,
				"X" : z.x,
				"Y" : z.y,
				"Z" : z.z
		};
//		if( g.$GalleryAll ) {
//			if( all ) {
//				ch += {"ZZEnc" : "Encounters"};
//				msg += "\nSelect \"Encounters\" to see Gallery of encounters.";
//			} else {
//				ch += {"ZZAll" : "All"};
//				msg += "\nSelect \"All\" to see Gallery of all objects.";
//			}
//		}
		if( isValidFrameCallback( g.$GalleryFCB ) ) removeFrameCallback( g.$GalleryFCB );
		mission.runScreen({
			title: title,
			message: msg,
			model: "["+player.ship.dataKey+"]",
			spinModel:false,
			background: "gallery_bg.png",
//			initialChoicesKey: "_"+g.$GalleryLastMenu,
			choices: ch
		},function(text) {
			if( text == "_" ) { //Last encounters menu
				g.$Gallery_LastEncounters( all, init );
				return;
			}
//			if( text == "__" ) { //exhibitions menu
//				var e = worldScripts["exhibitions"];
//				e.$ExhibitionsLastMenu = 0;
//				if( g.$GalleryLog) log("Gallery", "exhibitions menu");
//				if( !isValidFrameCallback( e.$ExhibitionsMenuFCB ) )
//					e.$ExhibitionsMenuFCB = addFrameCallback( e.$Exhibitions_MenuFCB ); //will call ex.menu
//			} else { //selected a ship
				g.$GalleryRole = 0;
//			if( g.$GalleryAll && !g.$GalleryShowAll && text == "ZZAll") {
//				g.$GalleryShowAll = true;
//				g.$GalleryKeysForRole[ 0 ] = g.$GalleryAllObjects; //fill up default key array
//			} else if( g.$GalleryAll && g.$GalleryShowAll && text == "ZZEnc") {
//				g.$GalleryShowAll = false;
//				g.$GalleryKeysForRole[ 0 ] = g.$GalleryEncounters; //fill up default key array
//			} else 
				g.$Gallery_SearchKey(text.toLowerCase(), all);
//			}
		});
		var m = mission.displayModel;
		if( m ) {
			var w = oolite.gameSettings.gameWindow; //correction if not in widescreen
			var wide = Math.max( 0.01, ( w.width / w.height ) / g.$GalleryDefaultZoom );
			m.position = Vector3D(m.position.x, m.position.y, m.position.z * wide );
			m.orientation = m.orientation.rotateY( Math.PI );
		}
	} else { //textEntry need Oolite v1.79
		msg += "\nType any part of a ship name or press enter for a list.\nDetails of your ship:\n\n";
//		if( g.$GalleryAll ) {
//			if( all ) msg += "\nType \"enc\" to switch back to the Gallery of encounters.";
//			else msg += "\nType \"all\" to switch to the Gallery of all objects.";
//		}
		msg += g.Gallery_GetShipData(player.ship, g.$GalleryShowAll);
		if( isValidFrameCallback( worldScripts["gallery"].$GalleryFCB ) )
			removeFrameCallback( worldScripts["gallery"].$GalleryFCB );
		mission.runScreen({
			title: title,
			message: msg,
			background: "gallery_bg.png",
			model: "["+player.ship.dataKey+"]",
			spinModel:true,
			screenID: "gallery-search",
			textEntry: true
		},function(text) {
			g.$GalleryRole = 0;
			if( !text || text.length == 0 ) {
				g.$Gallery_Search( all, true );
				return;
			}
			text = text.toLowerCase();
//			if( g.$GalleryAll && !g.$GalleryShowAll && text == "all") {
//				g.$GalleryShowAll = true;
//				g.$GalleryKeysForRole[ 0 ] = g.$GalleryAllObjects; //fill up default key array
//			} else if( g.$GalleryAll && g.$GalleryShowAll && text == "enc") {
//				g.$GalleryShowAll = false;
//				g.$GalleryKeysForRole[ 0 ] = g.$GalleryEncounters; //fill up default key array
//			} else 
				g.$Gallery_SearchKey(text, all);
		});
	}
}

this.$Gallery_SearchHeadMsg = function( all ) {
	var g = worldScripts["gallery"];
	var len = g.$GalleryEncounters.length;
	if( all ) len = len + " of "+g.$GalleryAllObjects.length;
	return( "Your "+player.ship.displayName+" has recorded "+len+" ship and space object types." );
}

this.$Gallery_SearchHeadTitle = function( all ) {
	var title = "Gallery of ";
	if( all ) title += "all objects";
	else title += "encounters";
	return( title );
}

this.$Gallery_SearchKey = function( text, all ) {
	var g = worldScripts["gallery"];
	var k = g.$GalleryKeysForRole[ 0 ];
	var found = false;
	if( !text || text.length == 0 ) found = true; //no search, show the last item
	else if( k ) {
		if( !g.$GalleryShowAll ) { //search in name, not in dataKey
			var n = g.$GalleryEncNames; 
			if(g.$GalleryLog) log("Gallery",n);
			for( var i = 0; i < n.length; i++ ) {
				if ( n[i] && n[i].toLowerCase().indexOf(text) == 0 ) { //find text from begin
					found = true;
					g.$GalleryKeyi[ 0 ] = i;//set to the found item
					i = n.length; //exit
				}
			}
			if( !found ) {
				for( var i = 0; i < n.length; i++ ) {
					if ( n[i] && n[i].toLowerCase().indexOf(text) != -1 ) { //find text within
						found = true;
						g.$GalleryKeyi[ 0 ] = i;//set to the found item
						i = n.length; //exit
					}
				}
			}
		}
		if( !found ) for( var i = 0; i < k.length; i++ ) { //search in dataKey
			if ( k[i] && k[i].toLowerCase().indexOf(text) == 0 ) { //find text from begin
				found = true;
				g.$GalleryKeyi[ 0 ] = i;//set to the found item
				i = k.length; //exit
			}
		}
		if( !found ) {
			for( var i = 0; i < k.length; i++ ) {
				if ( k[i] && k[i].toLowerCase().indexOf(text) != -1 ) { //find text within
					found = true;
					g.$GalleryKeyi[ 0 ] = i;//set to the found item
					i = k.length; //exit
				}
			}
		}
		if( !found && text.length > 1 ) {
			var text1 = text.charAt(0); //find starting letter only
			for( var i = 0; i < k.length; i++ ) {
				if ( k[i] && k[i].charAt(0).toLowerCase() == text1 ) {
					found = true;
					g.$GalleryKeyi[ 0 ] = i;//set to the found item
					i = k.length; //exit
				}
			}
		}
	}
	if( found ) g.$Gallery_Interface2( all ); //show it
	else g.$Gallery_Search( all, false );//try again
}

/*
this.$Gallery_Timed = function() { //extra ship.remove can not help avoid "Universe is full" due to the remaining entities
	for( var i = 0; i < g.$GalleryTrash.length; i++ ) {
		g.$GalleryTrash[i].remove(true);
	}
	if( g.$GalleryLog) log("Gallery", "Removed "+g.$GalleryTrash.length+" tried ship.");
	delete g.$GalleryTrash;
}
*/	

this.$Gallery_Timed = function() { //cause extreme memory usage and crash when out of allocation space
	var g = worldScripts["gallery"];
	var maxk = 10; //how many iteration max., very slow and cause "Universe is full" if many OXP ships
	var num = 64; //how many ships created at one time
	var i = g.$GalleryTRole;//actual role in this timed function
	var k = g.$GalleryTRI;//actual TRole Iteration
	var prevlen = 0;

	if( g.$GalleryTimer ) { //remove old timer
		g.$GalleryTimer.stop();
		delete g.$GalleryTimer;
	}

	if( g.$GalleryKeysForRole[ i ] ) prevlen = g.$GalleryKeysForRole[ i ].length;
	var ships = system.addShips(g.$GalleryRoles[i], num, [0,0,0], 1000000);
	if(ships) {
		for( var j = 0; j < ships.length; j++ ) { //created ships
			if( ships[j] ) {
				var key = ships[j].dataKey;
				if( !g.$GalleryKeysForRole[ i ] )
					g.$GalleryKeysForRole[ i ] = [key];
				else if( g.$GalleryKeysForRole[ i ].indexOf(key) == -1 )
					g.$GalleryKeysForRole[ i ].push(key);//add new key
				ships[j].remove(true);//can not remove everything, can reach Universe limit
//				if(g.$GalleryLog) log("Gallery", i + ". main role: " + g.$GalleryRoles[i]
//					+" "+k+". try, "+(j+1)+". key: "+key
//					+" GalleryKeysForRole["+i+"]:"+g.$GalleryKeysForRole[ i ]);
			}
		}
		if(g.$GalleryLog) log("Gallery", i + ". main role: " + g.$GalleryRoles[i]
			+" "+k+". try: "+g.$GalleryKeysForRole[ i ].length+" object." );
		if(prevlen == g.$GalleryKeysForRole[ i ].length) maxk = 0;//exit from iteration cycle
	}
	var end = false;
	g.$GalleryTRI++;
	if( g.$GalleryTRI > maxk ) {
	
		if(g.$GalleryKeysForRole[ i ]) {
			g.$GalleryKeysForRole[ i ].sort();
			if(g.$GalleryLog) log("Gallery", i + ". role after "+k+" try: " + g.$GalleryRoles[i]
				+ " ("+g.$GalleryKeysForRole[ i ].length+"): " + g.$GalleryKeysForRole[ i ] );
			g.$GalleryKeysForRole[ 0 ] = g.$GalleryKeysForRole[ 0 ].concat(g.$GalleryKeysForRole[ i ]);
		}

		g.$GalleryTRI = 1;
		g.$GalleryTRole++;
		if( g.$GalleryTRole >= g.$GalleryRoles.length ) {
			
			g.$GalleryKeysForRole[ 0 ].sort();
			for( var i = 1; i < g.$GalleryKeysForRole[ 0 ].length; i++ ) { //remove duplicated keys
				if(g.$GalleryKeysForRole[0][i] == g.$GalleryKeysForRole[0][i-1]) {
					g.$GalleryKeysForRole[0].splice(i,1); //remove this item
					i--;
				}
			}
			if(g.$GalleryLog) log("Gallery", "All dataKeys ("+g.$GalleryKeysForRole[ 0 ].length
				+"): " + g.$GalleryKeysForRole[ 0 ] );
			end = true; //dataKey search finished, no more timer start
		}
	}

	if( !end ) g.$GalleryTimer = new Timer(g, g.$Gallery_Timed, 2.3); //will make the next step
}


this.$GalleryValidateTarget = function(target) {

	if( target.dataKey.indexOf("stealth") > -1 || target.primaryRole.indexOf("stealth") > -1
		|| target.scriptInfo && target.scriptInfo.ccl_missionShip //skip mission ships
		|| target.name == "Constrictor"
		|| target.dataKey == "vector_arn" //mission ship in Vector OXP
		|| target.primaryRole.indexOf("rescue_blackbox") > -1 ) //mission ships in Rescue Stations OXP
		return false;
	
	if( !target.roles ) return true; //bugfix for entities without any role
		
	var lib = worldScripts["tech_ref_lib"];
	if( lib ) return( lib.$validateTarget(target) ); //call the original function if available
	
	//following lines comes from Technical Reference Library OXP by spara
	//check scanclasses
	if (this.$GalleryClassifiedScanClasses.indexOf(target.scanClass) != -1) return false;
	//check roles
	var targetRoles = target.roles;
	var i;
	for (i = 0; i < targetRoles.length; i++) {
		if (this.$GalleryClassifiedRoles.indexOf(targetRoles[i]) != -1) {
			return false;
		}
	}
	//check for script_info key
	if (target.scriptInfo && target.scriptInfo.classifiedShip) return false;
	return true;
}
